@@ -32,9 +32,9 @@ using DirectoryEntriesMap =
     AsyncMapConsumer<ModuleName, FileRoot::DirectoryEntries>;
 
 auto CreateDirectoryEntriesMap(
-    gsl::not_null<RepositoryConfig*> const& repo_config,
+    gsl::not_null<const RepositoryConfig*> const& repo_config,
     std::size_t jobs = 0) -> DirectoryEntriesMap;
 
 }  // namespace BuildMaps::Base
 
-#endif
+#endif  // INCLUDED_SRC_BUILDTOOL_BUILD_ENGINE_BASE_MAPS_DIRECTORY_MAP_HPP
