@@ -17,6 +17,6 @@
 #include <grpc/grpc.h>
 
 // Max size for batch transfers
-static constexpr std::size_t kMaxBatchTransferSize = 3 * 1024 * 1024;
+static constexpr std::size_t kMaxBatchTransferSize = 3UL * 1024 * 1024;
 static_assert(kMaxBatchTransferSize < GRPC_DEFAULT_MAX_RECV_MESSAGE_LENGTH,
               "Max batch transfer size too large.");
