@@ -18,7 +18,6 @@
 #include <algorithm>
 #include <atomic>
 #include <cstddef>
-#include <cstdint>
 #include <map>
 #include <memory>
 #include <optional>
@@ -38,6 +37,7 @@
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
 #include "src/utils/cpp/hex_string.hpp"
+#include "src/utils/cpp/transformed_range.hpp"
 #include "src/utils/cpp/type_safe_arithmetic.hpp"
 
 /// \brief Plain DirectedAcyclicGraph.
@@ -47,25 +47,23 @@
 class DirectedAcyclicGraph {
   public:
     /// \brief Abstract class for DAG nodes.
-    /// \tparam T_Content  Type of content.
-    /// \tparam T_Other    Type of neighboring nodes.
-    /// Sub classes need to implement \ref IsValid method.
+    /// \tparam TContent  Type of content.
+    /// \tparam TOther    Type of neighboring nodes.
     /// TODO: once we have hashes, require sub classes to implement generating
     /// IDs depending on its unique content.
-    template <typename T_Content, typename T_Other>
+    template <typename TContent, typename TOther>
     class Node {
       public:
-        using Id = std::uintptr_t;
-        using OtherNode = T_Other;
+        using OtherNode = TOther;
         using OtherNodePtr = gsl::not_null<OtherNode*>;
 
-        explicit Node(T_Content&& content) noexcept
+        explicit Node(TContent&& content) noexcept
             : content_{std::move(content)} {}
 
         // NOLINTNEXTLINE(modernize-pass-by-value)
-        explicit Node(T_Content const& content) noexcept : content_{content} {}
+        explicit Node(TContent const& content) noexcept : content_{content} {}
 
-        Node(T_Content const& content,
+        Node(TContent const& content,
              std::vector<OtherNodePtr> const& parents,
              std::vector<OtherNodePtr> const& children) noexcept
             : content_{content}, parents_{parents}, children_{children} {}
@@ -76,17 +74,13 @@ class DirectedAcyclicGraph {
         Node(Node&&) = delete;
         auto operator=(Node const&) -> Node& = delete;
         auto operator=(Node&&) -> Node& = delete;
+        ~Node() = default;
 
-        [[nodiscard]] auto NodeId() const noexcept -> Id {
-            // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
-            return reinterpret_cast<Id>(this);
-        }
-
-        [[nodiscard]] auto Content() const& noexcept -> T_Content const& {
+        [[nodiscard]] auto Content() const& noexcept -> TContent const& {
             return content_;
         }
 
-        [[nodiscard]] auto Content() && noexcept -> T_Content {
+        [[nodiscard]] auto Content() && noexcept -> TContent {
             return std::move(content_);
         }
 
@@ -120,11 +114,8 @@ class DirectedAcyclicGraph {
             return true;
         }
 
-        [[nodiscard]] virtual auto IsValid() const noexcept -> bool = 0;
-        virtual ~Node() noexcept = default;
-
       private:
-        T_Content content_{};
+        TContent content_{};
         std::vector<OtherNodePtr> parents_{};
         std::vector<OtherNodePtr> children_{};
     };
@@ -144,8 +135,8 @@ class DirectedAcyclicGraph {
         NodeTraversalState() noexcept = default;
         NodeTraversalState(NodeTraversalState const&) = delete;
         NodeTraversalState(NodeTraversalState&&) = delete;
-        auto operator=(NodeTraversalState const&)
-            -> NodeTraversalState& = delete;
+        auto operator=(NodeTraversalState const&) -> NodeTraversalState& =
+                                                         delete;
         auto operator=(NodeTraversalState&&) -> NodeTraversalState& = delete;
         ~NodeTraversalState() noexcept = default;
 
@@ -179,37 +170,6 @@ class DirectedAcyclicGraph {
         std::atomic<bool> is_queued_to_be_processed_{false};
         std::atomic<bool> is_required_{false};
     };
-
-  protected:
-    template <typename T_Node>
-    // NOLINTNEXTLINE(misc-no-recursion)
-    [[nodiscard]] static auto check_validity(
-        gsl::not_null<T_Node*> node) noexcept -> bool {
-        // Check node-specified validity (e.g. bipartiteness requirements)
-        if (!node->IsValid()) {
-            return false;
-        }
-
-        // Search for cycles
-        thread_local std::vector<typename T_Node::Id> stack{};
-        for (auto const& child : node->Children()) {
-            auto node_id = child->NodeId();
-
-            if (std::find(stack.begin(), stack.end(), node_id) != stack.end()) {
-                return false;
-            }
-
-            stack.push_back(node_id);
-            bool valid = check_validity(child);
-            stack.pop_back();
-
-            if (!valid) {
-                return false;
-            }
-        }
-
-        return true;
-    }
 };
 
 class DependencyGraph : DirectedAcyclicGraph {
@@ -218,13 +178,13 @@ class DependencyGraph : DirectedAcyclicGraph {
     class ArtifactNode;
 
     // Node identifier for actions
-    struct ActionNodeIdentifierTag : type_safe_arithmetic_tag<std::size_t> {};
-    using ActionNodeIdentifier = type_safe_arithmetic<ActionNodeIdentifierTag>;
+    struct ActionNodeIdentifierTag : TypeSafeArithmeticTag<std::size_t> {};
+    using ActionNodeIdentifier = TypeSafeArithmetic<ActionNodeIdentifierTag>;
 
     // Node identifier for artifacts
-    struct ArtifactNodeIdentifierTag : type_safe_arithmetic_tag<std::size_t> {};
+    struct ArtifactNodeIdentifierTag : TypeSafeArithmeticTag<std::size_t> {};
     using ArtifactNodeIdentifier =
-        type_safe_arithmetic<ArtifactNodeIdentifierTag>;
+        TypeSafeArithmetic<ArtifactNodeIdentifierTag>;
 
     /// \brief Class for traversal state data specific for ActionNode's
     /// Provides the following atomic operations (listed on the public methods):
@@ -292,7 +252,7 @@ class DependencyGraph : DirectedAcyclicGraph {
     };
 
     /// \brief Action node (bipartite)
-    /// Cannot be entry (see \ref IsValid).
+    /// Cannot be entry.
     class ActionNode final : public Node<Action, ArtifactNode> {
         using base = Node<Action, ArtifactNode>;
 
@@ -303,6 +263,9 @@ class DependencyGraph : DirectedAcyclicGraph {
             Action::LocalPath path;
             base::OtherNodePtr node;
         };
+        using LocalPaths =
+            TransformedRange<std::vector<NamedOtherNodePtr>::const_iterator,
+                             Action::LocalPath>;
 
         [[nodiscard]] static auto Create(Action const& content) noexcept
             -> Ptr {
@@ -313,11 +276,6 @@ class DependencyGraph : DirectedAcyclicGraph {
             return std::make_unique<ActionNode>(std::move(content));
         }
 
-        // only valid if it has parents
-        [[nodiscard]] auto IsValid() const noexcept -> bool final {
-            return (!base::Parents().empty());
-        }
-
         [[nodiscard]] auto AddOutputFile(
             NamedOtherNodePtr const& output) noexcept -> bool {
             base::AddParent(output.node);
@@ -360,74 +318,59 @@ class DependencyGraph : DirectedAcyclicGraph {
             return true;
         }
 
-        [[nodiscard]] auto OutputFiles()
-            const& -> std::vector<NamedOtherNodePtr> const& {
+        [[nodiscard]] auto OutputFiles() const& noexcept
+            -> std::vector<NamedOtherNodePtr> const& {
             return output_files_;
         }
 
-        [[nodiscard]] auto OutputDirs()
-            const& -> std::vector<NamedOtherNodePtr> const& {
+        [[nodiscard]] auto OutputDirs() const& noexcept
+            -> std::vector<NamedOtherNodePtr> const& {
             return output_dirs_;
         }
 
-        [[nodiscard]] auto Command() const -> std::vector<std::string> {
-            return Content().Command();
-        }
-
-        [[nodiscard]] auto Env() const -> std::map<std::string, std::string> {
-            return Content().Env();
-        }
-
-        [[nodiscard]] auto MayFail() const -> std::optional<std::string> {
-            return Content().MayFail();
-        }
-
-        [[nodiscard]] auto TimeoutScale() const -> double {
-            return Content().TimeoutScale();
-        }
-
-        [[nodiscard]] auto ExecutionProperties() const
-            -> std::map<std::string, std::string> {
-            return Content().ExecutionProperties();
-        }
-
-        [[nodiscard]] auto NoCache() const -> bool {
-            return Content().NoCache();
-        }
-
-        [[nodiscard]] auto Dependencies()
-            const& -> std::vector<NamedOtherNodePtr> const& {
+        [[nodiscard]] auto Dependencies() const& noexcept
+            -> std::vector<NamedOtherNodePtr> const& {
             return dependencies_;
         }
 
-        [[nodiscard]] auto OutputFilePaths() const
-            -> std::vector<Action::LocalPath> {
-            return NodePaths(output_files_);
+        [[nodiscard]] auto Command() const noexcept
+            -> std::vector<std::string> const& {
+            return Content().Command();
         }
 
-        [[nodiscard]] auto OutputFileIds() const
-            -> std::vector<ArtifactIdentifier> {
-            return Ids(output_files_);
+        [[nodiscard]] auto Env() const noexcept
+            -> std::map<std::string, std::string> const& {
+            return Content().Env();
         }
 
-        [[nodiscard]] auto OutputDirPaths() const
-            -> std::vector<Action::LocalPath> {
-            return NodePaths(output_dirs_);
+        [[nodiscard]] auto MayFail() const noexcept
+            -> std::optional<std::string> const& {
+            return Content().MayFail();
         }
 
-        [[nodiscard]] auto OutputDirIds() const
-            -> std::vector<ArtifactIdentifier> {
-            return Ids(output_dirs_);
+        [[nodiscard]] auto TimeoutScale() const noexcept -> double {
+            return Content().TimeoutScale();
         }
 
-        [[nodiscard]] auto DependencyPaths() const
-            -> std::vector<Action::LocalPath> {
-            return NodePaths(dependencies_);
+        [[nodiscard]] auto ExecutionProperties() const noexcept
+            -> std::map<std::string, std::string> const& {
+            return Content().ExecutionProperties();
         }
 
-        [[nodiscard]] auto DependencyIds() const
-            -> std::vector<ArtifactIdentifier> {
-            return Ids(dependencies_);
+        [[nodiscard]] auto NoCache() const noexcept -> bool {
+            return Content().NoCache();
+        }
+
+        [[nodiscard]] auto OutputFilePaths() const& noexcept -> LocalPaths {
+            return NodePaths(&output_files_);
+        }
+
+        [[nodiscard]] auto OutputDirPaths() const& noexcept -> LocalPaths {
+            return NodePaths(&output_dirs_);
+        }
+
+        [[nodiscard]] auto DependencyPaths() const& noexcept -> LocalPaths {
+            return NodePaths(&dependencies_);
         }
 
         // To initialise the action traversal specific data before traversing
@@ -448,39 +391,18 @@ class DependencyGraph : DirectedAcyclicGraph {
         std::unique_ptr<ActionNodeTraversalState> traversal_state_{
             std::make_unique<ActionNodeTraversalState>()};
 
-        // Collect paths from named nodes.
-        // TODO(oreiche): This could be potentially speed up by using a wrapper
-        // iterator to provide a read-only view (similar to BlobContainer)
         [[nodiscard]] static auto NodePaths(
-            std::vector<NamedOtherNodePtr> const& nodes)
-            -> std::vector<Action::LocalPath> {
-            std::vector<Action::LocalPath> paths{nodes.size()};
-            std::transform(
-                nodes.cbegin(),
-                nodes.cend(),
-                paths.begin(),
-                [](auto const& named_node) { return named_node.path; });
-            return paths;
-        }
-
-        /// \brief Collect ids from named nodes (artifacts in this case)
-        [[nodiscard]] static auto Ids(
-            std::vector<NamedOtherNodePtr> const& nodes)
-            -> std::vector<ArtifactIdentifier> {
-            std::vector<ArtifactIdentifier> ids{nodes.size()};
-            std::transform(nodes.cbegin(),
-                           nodes.cend(),
-                           ids.begin(),
-                           [](auto const& named_node) {
-                               return named_node.node->Content().Id();
-                           });
-            return ids;
+            gsl::not_null<std::vector<NamedOtherNodePtr> const*> const& nodes)
+            -> LocalPaths {
+            return TransformedRange{
+                nodes->begin(),
+                nodes->end(),
+                [](NamedOtherNodePtr const& node) { return node.path; }};
         }
     };
 
     /// \brief Artifact node (bipartite)
-    /// Can be entry or leaf (see \ref IsValid) and can only have single child
-    /// (see \ref AddChild)
+    /// Can be entry or leaf and can only have single child (see \ref AddChild)
     class ArtifactNode final : public Node<Artifact, ActionNode> {
         using base = Node<Artifact, ActionNode>;
 
@@ -515,12 +437,8 @@ class DependencyGraph : DirectedAcyclicGraph {
             return base::AddParent(action);
         }
 
-        [[nodiscard]] auto IsValid() const noexcept -> bool final {
-            return base::Children().size() <= 1;
-        }
-
         [[nodiscard]] auto HasBuilderAction() const noexcept -> bool {
-            return !base::Children().empty();
+            return not base::Children().empty();
         }
 
         [[nodiscard]] auto BuilderActionNode() const noexcept
@@ -570,47 +488,19 @@ class DependencyGraph : DirectedAcyclicGraph {
     [[nodiscard]] auto ActionNodeWithId(
         ActionIdentifier const& id) const noexcept -> ActionNode const*;
 
-    [[nodiscard]] auto ActionNodeOfArtifactWithId(
-        ArtifactIdentifier const& artifact_id) const noexcept
-        -> ActionNode const*;
-
-    [[nodiscard]] auto ArtifactWithId(
-        ArtifactIdentifier const& id) const noexcept -> std::optional<Artifact>;
-
-    [[nodiscard]] auto ActionWithId(ActionIdentifier const& id) const noexcept
-        -> std::optional<Action>;
-
-    [[nodiscard]] auto ActionOfArtifactWithId(
-        ArtifactIdentifier const& artifact_id) const noexcept
-        -> std::optional<Action>;
-
-    [[nodiscard]] auto ActionIdOfArtifactWithId(
-        ArtifactIdentifier const& artifact_id) const noexcept
-        -> std::optional<ActionIdentifier>;
-
-    [[nodiscard]] auto IsValid() const noexcept -> bool {
-        return std::all_of(
-            artifact_nodes_.begin(),
-            artifact_nodes_.end(),
-            [](auto const& node) {
-                return DirectedAcyclicGraph::check_validity<
-                    std::remove_reference_t<decltype(*node)>>(&(*node));
-            });
-    }
-
   private:
     // List of action nodes we already created
-    std::vector<ActionNode::Ptr> action_nodes_{};
+    std::vector<ActionNode::Ptr> action_nodes_;
 
     // List of artifact nodes we already created
-    std::vector<ArtifactNode::Ptr> artifact_nodes_{};
+    std::vector<ArtifactNode::Ptr> artifact_nodes_;
 
     // Associates global action identifier to local node id
-    std::unordered_map<ActionIdentifier, ActionNodeIdentifier> action_ids_{};
+    std::unordered_map<ActionIdentifier, ActionNodeIdentifier> action_ids_;
 
     // Associates global artifact identifier to local node id
     std::unordered_map<ArtifactIdentifier, ArtifactNodeIdentifier>
-        artifact_ids_{};
+        artifact_ids_;
 
     [[nodiscard]] auto CreateOutputArtifactNodes(
         std::string const& action_id,
