@@ -16,33 +16,43 @@
 #define INCLUDED_SRC_BUILDTOOL_FILE_SYSTEM_FILE_SYSTEM_MANAGER_HPP
 
 #include <array>
+#include <cerrno>  // for errno
 #include <chrono>
 #include <cstddef>
-#include <cstdio>  // for std::fopen
+#include <cstdint>
+#include <cstdio>   // for std::fopen
+#include <cstdlib>  // std::exit, std::getenv
 #include <cstring>
 #include <exception>
 #include <filesystem>
 #include <fstream>
 #include <optional>
+#include <system_error>
 #include <unordered_set>
+#include <variant>
 
 #ifdef __unix__
 #include <fcntl.h>
+#include <pwd.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <unistd.h>
+#else
+#error "Non-unix is not supported yet"
 #endif
 
 #include "gsl/gsl"
+#include "nlohmann/json.hpp"
 #include "src/buildtool/file_system/object_type.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
 #include "src/buildtool/system/system.hpp"
+#include "src/utils/cpp/expected.hpp"
 #include "src/utils/cpp/path.hpp"
 
 namespace detail {
-static inline consteval auto BitWidth(int max_val) -> int {
+static consteval auto BitWidth(int max_val) -> int {
     constexpr int kBitsPerByte = 8;
     int i = sizeof(max_val) * kBitsPerByte;
     while ((i-- > 0) and (((max_val >> i) & 0x01) == 0x00)) {  // NOLINT
@@ -69,9 +79,9 @@ class FileSystemManager {
         auto operator=(DirectoryAnchor const&) -> DirectoryAnchor& = delete;
         auto operator=(DirectoryAnchor&&) -> DirectoryAnchor& = delete;
         ~DirectoryAnchor() noexcept {
-            if (!kRestorePath.empty()) {
+            if (not restore_path_.empty()) {
                 try {
-                    std::filesystem::current_path(kRestorePath);
+                    std::filesystem::current_path(restore_path_);
                 } catch (std::exception const& e) {
                     Logger::Log(LogLevel::Error, e.what());
                 }
@@ -79,14 +89,14 @@ class FileSystemManager {
         }
         [[nodiscard]] auto GetRestorePath() const noexcept
             -> std::filesystem::path const& {
-            return kRestorePath;
+            return restore_path_;
         }
 
       private:
-        std::filesystem::path const kRestorePath{};
+        std::filesystem::path const restore_path_;
 
         DirectoryAnchor()
-            : kRestorePath{FileSystemManager::GetCurrentDirectory()} {}
+            : restore_path_{FileSystemManager::GetCurrentDirectory()} {}
         DirectoryAnchor(DirectoryAnchor&&) = default;
     };
 
@@ -139,6 +149,26 @@ class FileSystemManager {
         return CreateFileImpl(file) == CreationStatus::Created;
     }
 
+    /// \brief Determine user home directory
+    [[nodiscard]] static auto GetUserHome() noexcept -> std::filesystem::path {
+        char const* root{nullptr};
+
+#ifdef __unix__
+        root = std::getenv("HOME");
+        if (root == nullptr) {
+            root = getpwuid(getuid())->pw_dir;
+        }
+#endif
+
+        if (root == nullptr) {
+            Logger::Log(LogLevel::Error,
+                        "Cannot determine user home directory.");
+            std::exit(EXIT_FAILURE);
+        }
+
+        return root;
+    }
+
     /// \brief We are POSIX-compliant, therefore we only care about the string
     /// value the symlinks points to, whether it exists or not, not the target
     /// type. As such, we don't distinguish directory or file targets. However,
@@ -192,29 +222,36 @@ class FileSystemManager {
         return false;
     }
 
+    /// \brief Try to create a hard link; return unit on success and a
+    /// std::error_condition describing the failure.
     [[nodiscard]] static auto CreateFileHardlink(
         std::filesystem::path const& file_path,
         std::filesystem::path const& link_path,
-        LogLevel log_failure_at = LogLevel::Error) noexcept -> bool {
-        try {
-            std::filesystem::create_hard_link(file_path, link_path);
-            return std::filesystem::is_regular_file(link_path);
-        } catch (std::exception const& e) {
-            Logger::Log(log_failure_at,
-                        "hard linking {} to {}\n{}",
-                        file_path.string(),
-                        link_path.string(),
-                        e.what());
-            return false;
+        LogLevel log_failure_at = LogLevel::Error) noexcept
+        -> expected<std::monostate, std::error_code> {
+        std::error_code ec{};
+        std::filesystem::create_hard_link(file_path, link_path, ec);
+        if (not ec) {
+            if (std::filesystem::is_regular_file(link_path)) {
+                return std::monostate{};
+            }
+            return unexpected(std::error_code{});
         }
+        Logger::Log(log_failure_at,
+                    "failed hard linking {} to {}: {}, {}",
+                    nlohmann::json(file_path.string()).dump(),
+                    nlohmann::json(link_path.string()).dump(),
+                    ec.value(),
+                    ec.message());
+        return unexpected(ec);
     }
 
     template <ObjectType kType, bool kSetEpochTime = false>
-    requires(IsFileObject(kType))
-        [[nodiscard]] static auto CreateFileHardlinkAs(
-            std::filesystem::path const& file_path,
-            std::filesystem::path const& link_path,
-            LogLevel log_failure_at = LogLevel::Error) noexcept -> bool {
+        requires(IsFileObject(kType))
+    [[nodiscard]] static auto CreateFileHardlinkAs(
+        std::filesystem::path const& file_path,
+        std::filesystem::path const& link_path,
+        LogLevel log_failure_at = LogLevel::Error) noexcept -> bool {
         // Set permissions first (permissions are a property of the file) so
         // that the created link has the correct permissions as soon as the link
         // creation is finished.
@@ -318,7 +355,8 @@ class FileSystemManager {
     template <ObjectType kType,
               bool kSetEpochTime = false,
               bool kSetWritable = false>
-    requires(IsFileObject(kType)) [[nodiscard]] static auto CopyFileAs(
+        requires(IsFileObject(kType))
+    [[nodiscard]] static auto CopyFileAs(
         std::filesystem::path const& src,
         std::filesystem::path const& dst,
         bool fd_less = false,
@@ -381,11 +419,11 @@ class FileSystemManager {
                             dst.string());
                 return false;
             }
-            std::filesystem::copy(src,
-                                  dst,
-                                  recursively
-                                      ? std::filesystem::copy_options::recursive
-                                      : std::filesystem::copy_options::none);
+            auto const opts =
+                std::filesystem::copy_options::copy_symlinks |
+                (recursively ? std::filesystem::copy_options::recursive
+                             : std::filesystem::copy_options::none);
+            std::filesystem::copy(src, dst, opts);
             return true;
         } catch (std::exception const& e) {
             Logger::Log(LogLevel::Error,
@@ -447,11 +485,11 @@ class FileSystemManager {
         std::filesystem::path const& file) noexcept -> bool {
         try {
             auto status = std::filesystem::symlink_status(file);
-            if (!std::filesystem::exists(status)) {
+            if (not std::filesystem::exists(status)) {
                 return true;
             }
-            if (!std::filesystem::is_regular_file(status) and
-                !std::filesystem::is_symlink(status)) {
+            if (not std::filesystem::is_regular_file(status) and
+                not std::filesystem::is_symlink(status)) {
                 return false;
             }
             return std::filesystem::remove(file);
@@ -469,10 +507,10 @@ class FileSystemManager {
         -> bool {
         try {
             auto status = std::filesystem::symlink_status(dir);
-            if (!std::filesystem::exists(status)) {
+            if (not std::filesystem::exists(status)) {
                 return true;
             }
-            if (!std::filesystem::is_directory(status)) {
+            if (not std::filesystem::is_directory(status)) {
                 return false;
             }
             if (recursively) {
@@ -546,7 +584,7 @@ class FileSystemManager {
         -> bool {
         try {
             auto const status = std::filesystem::symlink_status(file);
-            if (!std::filesystem::is_regular_file(status)) {
+            if (not std::filesystem::is_regular_file(status)) {
                 return false;
             }
         } catch (std::exception const& e) {
@@ -665,7 +703,7 @@ class FileSystemManager {
             std::ifstream file_reader(file.string(), std::ios::binary);
             if (file_reader.is_open()) {
                 auto ssize = gsl::narrow<std::streamsize>(chunk.size());
-                do {
+                while (file_reader.good()) {
                     file_reader.read(chunk.data(), ssize);
                     auto count = file_reader.gcount();
                     if (count == ssize) {
@@ -675,7 +713,7 @@ class FileSystemManager {
                         content +=
                             chunk.substr(0, gsl::narrow<std::size_t>(count));
                     }
-                } while (file_reader.good());
+                }
                 file_reader.close();
                 return content;
             }
@@ -893,10 +931,10 @@ class FileSystemManager {
     template <ObjectType kType,
               bool kSetEpochTime = false,
               bool kSetWritable = false>
-    requires(IsFileObject(kType))
-        [[nodiscard]] static auto WriteFileAs(std::string const& content,
-                                              std::filesystem::path const& file,
-                                              bool fd_less = false) noexcept
+        requires(IsFileObject(kType))
+    [[nodiscard]] static auto WriteFileAs(std::string const& content,
+                                          std::filesystem::path const& file,
+                                          bool fd_less = false) noexcept
         -> bool {
         return WriteFile(content, file, fd_less) and
                SetFilePermissions<kSetWritable>(file,
@@ -947,7 +985,7 @@ class FileSystemManager {
     }
 
   private:
-    enum class CreationStatus { Created, Exists, Failed };
+    enum class CreationStatus : std::uint8_t { Created, Exists, Failed };
 
     static constexpr std::size_t kChunkSize{256};
 
@@ -1044,7 +1082,7 @@ class FileSystemManager {
         }
         try {
             std::ofstream writer{file};
-            if (!writer.is_open()) {
+            if (not writer.is_open()) {
                 Logger::Log(
                     LogLevel::Error, "can not open file {}", file.string());
                 return false;
@@ -1128,10 +1166,10 @@ class FileSystemManager {
         std::filesystem::file_status const& status) noexcept -> bool {
         try {
             namespace fs = std::filesystem;
-            static constexpr auto exec_flags = fs::perms::owner_exec bitor
+            static constexpr auto kExecFlags = fs::perms::owner_exec bitor
                                                fs::perms::group_exec bitor
                                                fs::perms::others_exec;
-            auto exec_perms = status.permissions() bitand exec_flags;
+            auto exec_perms = status.permissions() bitand kExecFlags;
             return exec_perms != fs::perms::none;
         } catch (std::exception const& e) {
             Logger::Log(LogLevel::Error,
@@ -1147,14 +1185,14 @@ class FileSystemManager {
     /// Non-zero return values indicate errors, which can be decoded using
     /// \ref ErrorToString.
     class LowLevel {
-        static constexpr ssize_t kDefaultChunkSize = 1024 * 32;
+        static constexpr std::size_t kDefaultChunkSize = 1024UL * 32;
         static constexpr int kWriteFlags =
             O_WRONLY | O_CREAT | O_TRUNC;           // NOLINT
         static constexpr int kWritePerms =          // 644
             S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;  // NOLINT
 
       public:
-        template <ssize_t kChunkSize = kDefaultChunkSize>
+        template <std::size_t kChunkSize = kDefaultChunkSize>
         [[nodiscard]] static auto CopyFile(char const* src,
                                            char const* dst,
                                            bool skip_existing) noexcept -> int {
@@ -1184,7 +1222,7 @@ class FileSystemManager {
             while ((len = read(in.fd, buf.data(), buf.size())) > 0) {
                 ssize_t wlen{};
                 ssize_t written_len{};
-                while (written_len < len &&
+                while (written_len < len and
                        (wlen = write(out.fd,
                                      buf.data() + written_len,  // NOLINT
                                      len - written_len)) > 0) {
@@ -1200,18 +1238,19 @@ class FileSystemManager {
             return 0;
         }
 
-        template <ssize_t kChunkSize = kDefaultChunkSize>
+        template <std::size_t kChunkSize = kDefaultChunkSize>
         [[nodiscard]] static auto WriteFile(char const* content,
-                                            ssize_t size,
+                                            std::size_t size,
                                             char const* file) noexcept -> int {
             auto out = FdOpener{file, kWriteFlags, kWritePerms};
             if (out.fd == -1) {
                 return PackError(ERROR_OPEN_OUTPUT, errno);
             }
-            ssize_t pos{};
+            std::size_t pos = 0;
             while (pos < size) {
                 auto const write_len = std::min(kChunkSize, size - pos);
-                auto len = write(out.fd, content + pos, write_len);  // NOLINT
+                auto const len =
+                    write(out.fd, content + pos, write_len);  // NOLINT
                 if (len < 0) {
                     return PackError(ERROR_WRITE_OUTPUT, errno);
                 }
@@ -1250,7 +1289,7 @@ class FileSystemManager {
         }
 
       private:
-        enum ErrorCodes {
+        enum ErrorCodes : std::uint8_t {
             ERROR_READ_INPUT,    // read() input file failed
             ERROR_OPEN_INPUT,    // open() input file failed
             ERROR_OPEN_OUTPUT,   // open() output file failed
@@ -1300,6 +1339,6 @@ class FileSystemManager {
         }
 
     };  // class LowLevel
-};      // class FileSystemManager
+};  // class FileSystemManager
 
 #endif  // INCLUDED_SRC_BUILDTOOL_FILE_SYSTEM_FILE_SYSTEM_MANAGER_HPP
