@@ -21,16 +21,20 @@
 #include "src/buildtool/build_engine/base_maps/entity_name.hpp"
 #include "src/buildtool/build_engine/target_map/configured_target.hpp"
 #include "src/buildtool/common/repository_config.hpp"
+#include "src/buildtool/execution_api/common/api_bundle.hpp"
+#include "src/buildtool/serve_api/remote/serve_api.hpp"
 
 [[nodiscard]] auto DescribeTarget(
     BuildMaps::Target::ConfiguredTarget const& id,
-    gsl::not_null<RepositoryConfig*> const& repo_config,
+    gsl::not_null<const RepositoryConfig*> const& repo_config,
+    std::optional<ServeApi> const& serve,
+    ApiBundle const& apis,
     std::size_t jobs,
     bool print_json) -> int;
 
 [[nodiscard]] auto DescribeUserDefinedRule(
     BuildMaps::Base::EntityName const& rule_name,
-    gsl::not_null<RepositoryConfig*> const& repo_config,
+    gsl::not_null<const RepositoryConfig*> const& repo_config,
     std::size_t jobs,
     bool print_json) -> int;
 
