@@ -31,7 +31,7 @@
 #include "src/buildtool/build_engine/expression/configuration.hpp"
 #include "src/buildtool/build_engine/expression/expression.hpp"
 #include "src/buildtool/common/location.hpp"
-#include "src/buildtool/common/remote/retry_parameters.hpp"
+#include "src/buildtool/crypto/hash_function.hpp"
 #include "src/buildtool/file_system/file_system_manager.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
@@ -45,66 +45,54 @@ namespace {
     -> std::optional<std::pair<std::filesystem::path, std::filesystem::path>> {
     if (location.IsNotNull()) {
         auto res = ReadLocationObject(location->ToJson(), std::nullopt);
-        if (res.index() == 0) {
-            Logger::Log(LogLevel::Error, std::get<0>(res));
+        if (not res) {
+            Logger::Log(LogLevel::Error, res.error());
             std::exit(kExitFailure);
         }
-        return std::get<1>(res);
+        return *res;
     }
     return std::nullopt;
 }
 
 }  // namespace
 
-[[nodiscard]] auto ParseRetryCliOptions(Configuration const& config) noexcept
-    -> bool {
+[[nodiscard]] auto ParseRetryCliOptions(
+    Configuration const& config,
+    gsl::not_null<RetryArguments*> const& rargs) noexcept -> bool {
     auto max_attempts = config["max-attempts"];
     if (max_attempts.IsNotNull()) {
-        if (!max_attempts->IsNumber()) {
+        if (not max_attempts->IsNumber()) {
             Logger::Log(
                 LogLevel::Error,
                 "Invalid value for \"max-attempts\" {}. It must be a number.",
                 max_attempts->ToString());
             return false;
         }
-        if (!Retry::SetMaxAttempts(max_attempts->Number())) {
-            Logger::Log(LogLevel::Error,
-                        "Invalid value for \"max-attempts\" {}.",
-                        max_attempts->Number());
-            return false;
-        }
+        rargs->max_attempts = static_cast<unsigned int>(max_attempts->Number());
     }
     auto initial_backoff = config["initial-backoff-seconds"];
     if (initial_backoff.IsNotNull()) {
-        if (!initial_backoff->IsNumber()) {
+        if (not initial_backoff->IsNumber()) {
             Logger::Log(LogLevel::Error,
                         "Invalid value \"initial-backoff-seconds\" {}. It must "
                         "be a number.",
                         initial_backoff->ToString());
             return false;
         }
-        if (!Retry::SetMaxAttempts(initial_backoff->Number())) {
-            Logger::Log(LogLevel::Error,
-                        "Invalid value for \"initial-backoff-seconds\" {}.",
-                        initial_backoff->Number());
-            return false;
-        }
+        rargs->initial_backoff_seconds =
+            static_cast<unsigned int>(initial_backoff->Number());
     }
     auto max_backoff = config["max-backoff-seconds"];
     if (max_backoff.IsNotNull()) {
-        if (!max_backoff->IsNumber()) {
+        if (not max_backoff->IsNumber()) {
             Logger::Log(LogLevel::Error,
                         "Invalid value for \"max-backoff-seconds\" {}. It must "
                         "be a number.",
                         max_backoff->ToString());
             return false;
         }
-        if (!Retry::SetMaxAttempts(max_backoff->Number())) {
-            Logger::Log(LogLevel::Error,
-                        "Invalid value for \"max-backoff-seconds\" {}.",
-                        max_backoff->Number());
-            return false;
-        }
+        rargs->max_backoff_seconds =
+            static_cast<unsigned int>(max_backoff->Number());
     }
     return true;
 }
@@ -403,7 +391,7 @@ void ReadJustServeConfig(gsl::not_null<CommandLineArguments*> const& clargs) {
             }
             // compatibility is set immediately if flag is true
             if (compatible->Bool()) {
-                Compatibility::SetCompatible();
+                clargs->protocol.hash_type = HashFunction::Type::PlainSHA256;
             }
         }
         // read the address
@@ -420,7 +408,7 @@ void ReadJustServeConfig(gsl::not_null<CommandLineArguments*> const& clargs) {
             }
             clargs->endpoint.remote_execution_address = address->String();
         }
-        if (!ParseRetryCliOptions(serve_config)) {
+        if (not ParseRetryCliOptions(serve_config, &clargs->retry)) {
             std::exit(kExitFailure);
         }
     }
@@ -435,7 +423,7 @@ void ReadJustServeConfig(gsl::not_null<CommandLineArguments*> const& clargs) {
                         jobs->ToString());
             std::exit(kExitFailure);
         }
-        clargs->common.jobs = jobs->Number();
+        clargs->common.jobs = static_cast<std::size_t>(jobs->Number());
     }
     // read build options
     auto build_args = serve_config["build"];
@@ -460,7 +448,8 @@ void ReadJustServeConfig(gsl::not_null<CommandLineArguments*> const& clargs) {
                     build_jobs->ToString());
                 std::exit(kExitFailure);
             }
-            clargs->build.build_jobs = build_jobs->Number();
+            clargs->build.build_jobs =
+                static_cast<std::size_t>(build_jobs->Number());
         }
         else {
             clargs->build.build_jobs = clargs->common.jobs;
@@ -478,7 +467,8 @@ void ReadJustServeConfig(gsl::not_null<CommandLineArguments*> const& clargs) {
                 std::exit(kExitFailure);
             }
             clargs->build.timeout =
-                std::size_t(timeout->Number()) * std::chrono::seconds{1};
+                static_cast<std::size_t>(timeout->Number()) *
+                std::chrono::seconds{1};
         }
         // read target-cache writing strategy
         auto strategy = build_args->Get("target-cache write strategy",
