@@ -30,18 +30,16 @@ void config_iter_closer(gsl::owner<git_config_iterator*> iter) {
 }
 
 // callback to enable SSL certificate check for remote fetch
-const auto certificate_check_cb = [](git_cert* /*cert*/,
-                                     int /*valid*/,
-                                     const char* /*host*/,
-                                     void* /*payload*/) -> int { return 1; };
+const auto kCertificateCheck = [](git_cert* /*cert*/,
+                                  int /*valid*/,
+                                  const char* /*host*/,
+                                  void* /*payload*/) -> int { return 1; };
 
 // callback to remote fetch without an SSL certificate check
-const auto certificate_passthrough_cb = [](git_cert* /*cert*/,
-                                           int /*valid*/,
-                                           const char* /*host*/,
-                                           void* /*payload*/) -> int {
-    return 0;
-};
+const auto kCertificatePassthrough = [](git_cert* /*cert*/,
+                                        int /*valid*/,
+                                        const char* /*host*/,
+                                        void* /*payload*/) -> int { return 0; };
 
 /// \brief Custom comparison of matching degrees. Return true if left argument's
 /// degree of matching is better that the right argument's. When both are
@@ -187,8 +185,8 @@ auto GitConfigSettings::GetSSLCallback(std::shared_ptr<git_config> const& cfg,
             }
         }
         // set callback: passthrough only if check_cert is false
-        return (check_cert and not *check_cert) ? certificate_passthrough_cb
-                                                : certificate_check_cb;
+        return (check_cert and not *check_cert) ? kCertificatePassthrough
+                                                : kCertificateCheck;
     } catch (std::exception const& ex) {
         (*logger)(
             fmt::format("Getting SSL callback failed with:\n{}", ex.what()),
@@ -327,7 +325,7 @@ auto GitConfigSettings::GetProxySettings(std::shared_ptr<git_config> const& cfg,
                                     true /*fatal*/);
                                 return std::nullopt;
                             }
-                            return proxy_info.value();
+                            return proxy_info;
                         }
                     }
                     // check the generic "http.proxy" gitconfig entry;
@@ -348,7 +346,7 @@ auto GitConfigSettings::GetProxySettings(std::shared_ptr<git_config> const& cfg,
                                     true /*fatal*/);
                                 return std::nullopt;
                             }
-                            return proxy_info.value();
+                            return proxy_info;
                         }
                         // cleanup memory
                         git_buf_dispose(&tmp_buf);
@@ -376,7 +374,7 @@ auto GitConfigSettings::GetProxySettings(std::shared_ptr<git_config> const& cfg,
                                     true /*fatal*/);
                                 return std::nullopt;
                             }
-                            return proxy_info.value();
+                            return proxy_info;
                         }
                         // check HTTPS_PROXY envariable
                         if (const char* envar = std::getenv("HTTPS_PROXY")) {
@@ -390,7 +388,7 @@ auto GitConfigSettings::GetProxySettings(std::shared_ptr<git_config> const& cfg,
                                     true /*fatal*/);
                                 return std::nullopt;
                             }
-                            return proxy_info.value();
+                            return proxy_info;
                         }
                     }
                     else if (url_scheme.value() == "http") {
@@ -406,7 +404,7 @@ auto GitConfigSettings::GetProxySettings(std::shared_ptr<git_config> const& cfg,
                                     true /*fatal*/);
                                 return std::nullopt;
                             }
-                            return proxy_info.value();
+                            return proxy_info;
                         }
                     }
                     // check all_proxy envariable
@@ -421,7 +419,7 @@ auto GitConfigSettings::GetProxySettings(std::shared_ptr<git_config> const& cfg,
                                 true /*fatal*/);
                             return std::nullopt;
                         }
-                        return proxy_info.value();
+                        return proxy_info;
                     }
                     // check ALL_PROXY envariable
                     if (const char* envar = std::getenv("ALL_PROXY")) {
@@ -435,7 +433,7 @@ auto GitConfigSettings::GetProxySettings(std::shared_ptr<git_config> const& cfg,
                                 true /*fatal*/);
                             return std::nullopt;
                         }
-                        return proxy_info.value();
+                        return proxy_info;
                     }
                 }
             }
