@@ -18,6 +18,9 @@
 #include <filesystem>
 #include <optional>
 
+#include "src/buildtool/common/retry_cli.hpp"
+#include "src/buildtool/storage/config.hpp"
+#include "src/buildtool/storage/storage.hpp"
 #include "src/other_tools/just_mr/cli.hpp"
 
 /// \brief Runs execvp for configured command. Only returns if execvp fails.
@@ -30,7 +33,9 @@
     MultiRepoRemoteAuthArguments const& auth_args,
     RetryArguments const& retry_args,
     ForwardOnlyArguments const& launch_fwd,
+    StorageConfig const& storage_config,
+    Storage const& storage,
     bool forward_build_root,
-    std::string multi_repo_tool_name) -> int;
+    std::string const& multi_repo_tool_name) -> int;
 
 #endif  // INCLUDED_SRC_OTHER_TOOLS_JUST_MR_LAUNCH_HPP
