@@ -27,16 +27,17 @@
 #include "nlohmann/json.hpp"
 #include "src/buildtool/build_engine/expression/configuration.hpp"
 #include "src/buildtool/common/retry_cli.hpp"
-#include "src/buildtool/compatibility/compatibility.hpp"
+#include "src/buildtool/crypto/hash_function.hpp"
 #include "src/buildtool/file_system/git_context.hpp"
 #include "src/buildtool/logging/log_config.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/log_sink_cmdline.hpp"
 #include "src/buildtool/logging/log_sink_file.hpp"
 #include "src/buildtool/logging/logger.hpp"
-#include "src/buildtool/main/retry.hpp"
 #include "src/buildtool/main/version.hpp"
+#include "src/buildtool/storage/config.hpp"
 #include "src/buildtool/storage/garbage_collector.hpp"
+#include "src/buildtool/storage/repository_garbage_collector.hpp"
 #include "src/other_tools/just_mr/cli.hpp"
 #include "src/other_tools/just_mr/exit_codes.hpp"
 #include "src/other_tools/just_mr/fetch.hpp"
@@ -100,6 +101,8 @@ void SetupSetupCommandArguments(
         "Advance Git commit IDs and print updated just-mr configuration.");
     auto* cmd_do = app.add_subcommand(
         "do", "Canonical way of specifying subcommands to be launched.");
+    auto* cmd_gc_repo = app.add_subcommand(
+        "gc-repo", "Perform garbage collection on the repository roots.");
     cmd_do->set_help_flag();  // disable help flag
     // define just subcommands
     std::vector<CLI::App*> cmd_just_subcmds{};
@@ -132,8 +135,11 @@ void SetupSetupCommandArguments(
     try {
         app.parse(argc, argv);
     } catch (CLI::Error& e) {
-        [[maybe_unused]] auto err = app.exit(e);
-        std::exit(kExitClargsError);
+        // CLI11 throws for things like --help calls for them to be handled
+        // separately by parse callers. In this case it nevertheless sets the
+        // error code to 0 (success).
+        auto const err = app.exit(e);
+        std::exit(err == 0 ? kExitSuccess : kExitClargsError);
     } catch (std::exception const& ex) {
         Logger::Log(LogLevel::Error, "Command line parse error: {}", ex.what());
         std::exit(kExitClargsError);
@@ -154,6 +160,9 @@ void SetupSetupCommandArguments(
     else if (*cmd_update) {
         clargs.cmd = SubCommand::kUpdate;
     }
+    else if (*cmd_gc_repo) {
+        clargs.cmd = SubCommand::kGcRepo;
+    }
     else if (*cmd_do) {
         clargs.cmd = SubCommand::kJustDo;
         // get remaining args
@@ -197,6 +206,25 @@ void SetupLogging(MultiRepoLogArguments const& clargs) {
     }
 }
 
+[[nodiscard]] auto CreateStorageConfig(MultiRepoCommonArguments const& args,
+                                       HashFunction::Type hash_type) noexcept
+    -> std::optional<StorageConfig> {
+    StorageConfig::Builder builder;
+    if (args.just_mr_paths->root.has_value()) {
+        builder.SetBuildRoot(*args.just_mr_paths->root);
+    }
+    builder.SetHashType(hash_type);
+
+    // As just-mr does not require the TargetCache, we do not need to set any of
+    // the remote execution fields for the backend description.
+    auto config = builder.Build();
+    if (config) {
+        return *std::move(config);
+    }
+    Logger::Log(LogLevel::Error, config.error());
+    return std::nullopt;
+}
+
 }  // namespace
 
 auto main(int argc, char* argv[]) -> int {
@@ -208,6 +236,7 @@ auto main(int argc, char* argv[]) -> int {
             my_name = std::filesystem::path(argv[0]).filename().string();
         } catch (...) {
             // ignore, as my_name is only used for error messages
+            my_name.clear();
         }
     }
     try {
@@ -287,14 +316,25 @@ auto main(int argc, char* argv[]) -> int {
             arguments.common.explicit_distdirs.end());
 
         // Setup LocalStorageConfig to store the local_build_root properly
-        // and make the cas and git cache roots available
-        if (not StorageConfig::SetBuildRoot(
-                *arguments.common.just_mr_paths->root)) {
+        // and make the cas and git cache roots available. A native storage is
+        // always instantiated, while a compatible one only if needed.
+        auto const native_storage_config =
+            CreateStorageConfig(arguments.common, HashFunction::Type::GitSHA1);
+        if (not native_storage_config) {
             Logger::Log(LogLevel::Error,
                         "Failed to configure local build root.");
             return kExitGenericFailure;
         }
 
+        if (arguments.cmd == SubCommand::kGcRepo) {
+            return RepositoryGarbageCollector::TriggerGarbageCollection(
+                       *native_storage_config)
+                       ? kExitSuccess
+                       : kExitBuiltinCommandFailure;
+        }
+
+        auto const native_storage = Storage::Create(&*native_storage_config);
+
         // check for conflicts in main repo name
         if ((not arguments.setup.sub_all) and arguments.common.main and
             arguments.setup.sub_main and
@@ -314,10 +354,6 @@ auto main(int argc, char* argv[]) -> int {
             return kExitGenericFailure;
         }
 
-        if (not SetupRetryConfig(arguments.retry)) {
-            return kExitGenericFailure;
-        }
-
         /**
          * The current implementation of libgit2 uses pthread_key_t incorrectly
          * on POSIX systems to handle thread-specific data, which requires us to
@@ -329,14 +365,6 @@ auto main(int argc, char* argv[]) -> int {
         // Run subcommands known to just and `do`
         if (arguments.cmd == SubCommand::kJustDo or
             arguments.cmd == SubCommand::kJustSubCmd) {
-            // check setup configuration arguments for validity
-            if (arguments.common.compatible == true and
-                arguments.common.fetch_absent) {
-                Logger::Log(LogLevel::Error,
-                            "Fetching absent repositories only available in "
-                            "native mode!");
-                return kExitConfigError;
-            }
             return CallJust(config_file,
                             arguments.common,
                             arguments.setup,
@@ -345,10 +373,17 @@ auto main(int argc, char* argv[]) -> int {
                             arguments.auth,
                             arguments.retry,
                             arguments.launch_fwd,
+                            *native_storage_config,
+                            native_storage,
                             forward_build_root,
                             my_name);
         }
-        auto lock = GarbageCollector::SharedLock();
+        auto repo_lock =
+            RepositoryGarbageCollector::SharedLock(*native_storage_config);
+        if (not repo_lock) {
+            return kExitGenericFailure;
+        }
+        auto lock = GarbageCollector::SharedLock(*native_storage_config);
         if (not lock) {
             return kExitGenericFailure;
         }
@@ -360,20 +395,15 @@ auto main(int argc, char* argv[]) -> int {
         // Run subcommand `setup` or `setup-env`
         if (arguments.cmd == SubCommand::kSetup or
             arguments.cmd == SubCommand::kSetupEnv) {
-            // check setup configuration arguments for validity
-            if (arguments.common.compatible == true and
-                arguments.common.fetch_absent) {
-                Logger::Log(LogLevel::Error,
-                            "Fetching absent repositories only available in "
-                            "native mode!");
-                return kExitConfigError;
-            }
             auto mr_config_path = MultiRepoSetup(
                 config,
                 arguments.common,
                 arguments.setup,
                 arguments.just_cmd,
                 arguments.auth,
+                arguments.retry,
+                *native_storage_config,
+                native_storage,
                 /*interactive=*/(arguments.cmd == SubCommand::kSetupEnv),
                 my_name);
             // dump resulting config to stdout
@@ -389,34 +419,23 @@ auto main(int argc, char* argv[]) -> int {
 
         // Run subcommand `update`
         if (arguments.cmd == SubCommand::kUpdate) {
-            return MultiRepoUpdate(
-                config, arguments.common, arguments.update, my_name);
+            return MultiRepoUpdate(config,
+                                   arguments.common,
+                                   arguments.update,
+                                   *native_storage_config,
+                                   my_name);
         }
 
         // Run subcommand `fetch`
         if (arguments.cmd == SubCommand::kFetch) {
-            // check fetch configuration arguments for validity
-            if (arguments.common.compatible == true) {
-                if (arguments.common.remote_execution_address and
-                    arguments.fetch.backup_to_remote) {
-                    Logger::Log(
-                        LogLevel::Error,
-                        "Remote backup for fetched archives only available "
-                        "in native mode!");
-                    return kExitConfigError;
-                }
-                if (arguments.common.fetch_absent) {
-                    Logger::Log(LogLevel::Error,
-                                "Fetching absent repositories only available "
-                                "in native mode!");
-                    return kExitConfigError;
-                }
-            }
             return MultiRepoFetch(config,
                                   arguments.common,
                                   arguments.setup,
                                   arguments.fetch,
                                   arguments.auth,
+                                  arguments.retry,
+                                  *native_storage_config,
+                                  native_storage,
                                   my_name);
         }
 
