@@ -18,12 +18,15 @@
 #include <memory>
 
 #include "src/buildtool/build_engine/expression/configuration.hpp"
+#include "src/buildtool/storage/config.hpp"
 #include "src/other_tools/just_mr/cli.hpp"
 
 /// \brief Update of Git repos commit information for a multi-repository build.
 [[nodiscard]] auto MultiRepoUpdate(std::shared_ptr<Configuration> const& config,
                                    MultiRepoCommonArguments const& common_args,
                                    MultiRepoUpdateArguments const& update_args,
-                                   std::string multi_repo_tool_name) -> int;
+                                   StorageConfig const& native_storage_config,
+                                   std::string const& multi_repo_tool_name)
+    -> int;
 
 #endif  // INCLUDED_SRC_OTHER_TOOLS_JUST_MR_UPDATE_HPP
