@@ -12,6 +12,9 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+#ifndef INCLUDED_SRC_TEST_BUILDTOOL_EXECUTION_API_COMMON_API_TEST_HPP
+#define INCLUDED_SRC_TEST_BUILDTOOL_EXECUTION_API_COMMON_API_TEST_HPP
+
 #include <cstdlib>
 #include <filesystem>
 #include <functional>
@@ -20,7 +23,9 @@
 #include <vector>
 
 #include "catch2/catch_test_macros.hpp"
-#include "src/buildtool/common/artifact_factory.hpp"
+#include "src/buildtool/common/artifact_description.hpp"
+#include "src/buildtool/common/artifact_digest_factory.hpp"
+#include "src/buildtool/crypto/hash_function.hpp"
 #include "src/buildtool/execution_api/common/execution_action.hpp"
 #include "src/buildtool/execution_api/common/execution_api.hpp"
 #include "src/buildtool/execution_api/common/execution_response.hpp"
@@ -28,11 +33,13 @@
 #include "src/buildtool/file_system/file_system_manager.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
+#include "test/utils/hermeticity/test_hash_function_type.hpp"
 
 using ApiFactory = std::function<IExecutionApi::Ptr()>;
 using ExecProps = std::map<std::string, std::string>;
 
-inline void SetLauncher() {
+[[nodiscard]] static inline auto CreateLocalExecConfig() noexcept
+    -> LocalExecutionConfig {
     std::vector<std::string> launcher{"env"};
     auto* env_path = std::getenv("PATH");
     if (env_path != nullptr) {
@@ -41,10 +48,12 @@ inline void SetLauncher() {
     else {
         launcher.emplace_back("PATH=/bin:/usr/bin");
     }
-    if (not LocalExecutionConfig::SetLauncher(launcher)) {
-        Logger::Log(LogLevel::Error, "Failure setting the local launcher.");
-        std::exit(EXIT_FAILURE);
+    LocalExecutionConfig::Builder builder;
+    if (auto config = builder.SetLauncher(std::move(launcher)).Build()) {
+        return *std::move(config);
     }
+    Logger::Log(LogLevel::Error, "Failure setting the local launcher.");
+    std::exit(EXIT_FAILURE);
 }
 
 [[nodiscard]] static inline auto GetTestDir(std::string const& test_name)
@@ -65,10 +74,13 @@ inline void SetLauncher() {
 
     auto api = api_factory();
 
-    auto action = api->CreateAction(
-        *api->UploadTree({}), {"echo", "-n", test_content}, {}, {}, {}, props);
-
-    SetLauncher();
+    auto action = api->CreateAction(*api->UploadTree({}),
+                                    {"echo", "-n", test_content},
+                                    "",
+                                    {},
+                                    {},
+                                    {},
+                                    props);
 
     SECTION("Cache execution result in action cache") {
         action->SetCacheFlag(IExecutionAction::CacheFlag::CacheOutput);
@@ -127,7 +139,11 @@ inline void SetLauncher() {
     ExecProps const& props,
     bool is_hermetic = false) {
     std::string test_content("test");
-    auto test_digest = ArtifactDigest::Create<ObjectType::File>(test_content);
+
+    HashFunction const hash_function{TestHashType::ReadFromEnvironment()};
+
+    auto test_digest = ArtifactDigestFactory::HashDataAs<ObjectType::File>(
+        hash_function, test_content);
 
     std::string output_path{"output_file"};
 
@@ -138,6 +154,7 @@ inline void SetLauncher() {
         {"/bin/sh",
          "-c",
          "set -e\necho -n " + test_content + " > " + output_path},
+        "",
         {output_path},
         {},
         {},
@@ -147,26 +164,28 @@ inline void SetLauncher() {
         action->SetCacheFlag(IExecutionAction::CacheFlag::CacheOutput);
 
         // run execution
-        auto response = action->Execute();
+        auto const response = action->Execute();
         REQUIRE(response);
 
         // verify result
-        auto artifacts = response->Artifacts();
-        REQUIRE(artifacts.contains(output_path));
-        CHECK(artifacts.at(output_path).digest == test_digest);
+        auto const artifacts = response->Artifacts();
+        REQUIRE(artifacts.has_value());
+        REQUIRE(artifacts.value()->contains(output_path));
+        CHECK(artifacts.value()->at(output_path).digest == test_digest);
 
         if (is_hermetic) {
             CHECK(not response->IsCached());
 
             SECTION("Rerun execution to verify caching") {
                 // run execution
-                auto response = action->Execute();
+                auto const response = action->Execute();
                 REQUIRE(response);
 
                 // verify result
-                auto artifacts = response->Artifacts();
-                REQUIRE(artifacts.contains(output_path));
-                CHECK(artifacts.at(output_path).digest == test_digest);
+                auto const artifacts = response->Artifacts();
+                REQUIRE(artifacts.has_value());
+                REQUIRE(artifacts.value()->contains(output_path));
+                CHECK(artifacts.value()->at(output_path).digest == test_digest);
                 CHECK(response->IsCached());
             }
         }
@@ -176,24 +195,25 @@ inline void SetLauncher() {
         action->SetCacheFlag(IExecutionAction::CacheFlag::DoNotCacheOutput);
 
         // run execution
-        auto response = action->Execute();
+        auto const response = action->Execute();
         REQUIRE(response);
 
         // verify result
-        auto artifacts = response->Artifacts();
-        REQUIRE(artifacts.contains(output_path));
-        CHECK(artifacts.at(output_path).digest == test_digest);
+        auto const artifacts = response->Artifacts();
+        REQUIRE(artifacts.has_value());
+        REQUIRE(artifacts.value()->contains(output_path));
+        CHECK(artifacts.value()->at(output_path).digest == test_digest);
         CHECK(not response->IsCached());
 
         SECTION("Rerun execution to verify caching") {
             // run execution
-            auto response = action->Execute();
+            auto const response = action->Execute();
             REQUIRE(response);
 
             // verify result
-            auto artifacts = response->Artifacts();
-            REQUIRE(artifacts.contains(output_path));
-            CHECK(artifacts.at(output_path).digest == test_digest);
+            auto const artifacts = response->Artifacts();
+            REQUIRE(artifacts.value()->contains(output_path));
+            CHECK(artifacts.value()->at(output_path).digest == test_digest);
             CHECK(not response->IsCached());
         }
     }
@@ -204,26 +224,30 @@ inline void SetLauncher() {
     ExecProps const& props,
     bool is_hermetic = false) {
     std::string test_content("test");
-    auto test_digest = ArtifactDigest::Create<ObjectType::File>(test_content);
+
+    HashFunction const hash_function{TestHashType::ReadFromEnvironment()};
+
+    auto test_digest = ArtifactDigestFactory::HashDataAs<ObjectType::File>(
+        hash_function, test_content);
 
     auto input_artifact_opt =
-        ArtifactFactory::FromDescription(ArtifactFactory::DescribeKnownArtifact(
-            test_digest.hash(), test_digest.size(), ObjectType::File));
-    CHECK(input_artifact_opt.has_value());
+        ArtifactDescription::CreateKnown(test_digest, ObjectType::File)
+            .ToArtifact();
     auto input_artifact =
-        DependencyGraph::ArtifactNode{std::move(*input_artifact_opt)};
+        DependencyGraph::ArtifactNode{std::move(input_artifact_opt)};
 
     std::string input_path{"dir/subdir/input"};
     std::string output_path{"output_file"};
 
     auto api = api_factory();
-    CHECK(api->Upload(BlobContainer{{BazelBlob{
+    CHECK(api->Upload(ArtifactBlobContainer{{ArtifactBlob{
                           test_digest, test_content, /*is_exec=*/false}}},
                       false));
 
     auto action =
         api->CreateAction(*api->UploadTree({{input_path, &input_artifact}}),
                           {"cp", input_path, output_path},
+                          "",
                           {output_path},
                           {},
                           {},
@@ -233,26 +257,28 @@ inline void SetLauncher() {
         action->SetCacheFlag(IExecutionAction::CacheFlag::CacheOutput);
 
         // run execution
-        auto response = action->Execute();
+        auto const response = action->Execute();
         REQUIRE(response);
 
         // verify result
-        auto artifacts = response->Artifacts();
-        REQUIRE(artifacts.contains(output_path));
-        CHECK(artifacts.at(output_path).digest == test_digest);
+        auto const artifacts = response->Artifacts();
+        REQUIRE(artifacts.has_value());
+        REQUIRE(artifacts.value()->contains(output_path));
+        CHECK(artifacts.value()->at(output_path).digest == test_digest);
 
         if (is_hermetic) {
             CHECK(not response->IsCached());
 
             SECTION("Rerun execution to verify caching") {
                 // run execution
-                auto response = action->Execute();
+                auto const response = action->Execute();
                 REQUIRE(response);
 
                 // verify result
-                auto artifacts = response->Artifacts();
-                REQUIRE(artifacts.contains(output_path));
-                CHECK(artifacts.at(output_path).digest == test_digest);
+                auto const artifacts = response->Artifacts();
+                REQUIRE(artifacts.has_value());
+                REQUIRE(artifacts.value()->contains(output_path));
+                CHECK(artifacts.value()->at(output_path).digest == test_digest);
                 CHECK(response->IsCached());
             }
         }
@@ -262,24 +288,26 @@ inline void SetLauncher() {
         action->SetCacheFlag(IExecutionAction::CacheFlag::DoNotCacheOutput);
 
         // run execution
-        auto response = action->Execute();
+        auto const response = action->Execute();
         REQUIRE(response);
 
         // verify result
-        auto artifacts = response->Artifacts();
-        REQUIRE(artifacts.contains(output_path));
-        CHECK(artifacts.at(output_path).digest == test_digest);
+        auto const artifacts = response->Artifacts();
+        REQUIRE(artifacts.has_value());
+        REQUIRE(artifacts.value()->contains(output_path));
+        CHECK(artifacts.value()->at(output_path).digest == test_digest);
         CHECK(not response->IsCached());
 
         SECTION("Rerun execution to verify caching") {
             // run execution
-            auto response = action->Execute();
+            auto const response = action->Execute();
             REQUIRE(response);
 
             // verify result
-            auto artifacts = response->Artifacts();
-            REQUIRE(artifacts.contains(output_path));
-            CHECK(artifacts.at(output_path).digest == test_digest);
+            auto const artifacts = response->Artifacts();
+            REQUIRE(artifacts.has_value());
+            REQUIRE(artifacts.value()->contains(output_path));
+            CHECK(artifacts.value()->at(output_path).digest == test_digest);
             CHECK(not response->IsCached());
         }
     }
@@ -289,7 +317,11 @@ inline void SetLauncher() {
     ApiFactory const& api_factory,
     ExecProps const& props) {
     std::string test_content("test");
-    auto test_digest = ArtifactDigest::Create<ObjectType::File>(test_content);
+
+    HashFunction const hash_function{TestHashType::ReadFromEnvironment()};
+
+    auto test_digest = ArtifactDigestFactory::HashDataAs<ObjectType::File>(
+        hash_function, test_content);
 
     std::string output_path{"output_file"};
 
@@ -300,6 +332,7 @@ inline void SetLauncher() {
                                      "-c",
                                      "set -e\necho -n " + test_content + " > " +
                                          output_path + "\nexit 1\n"},
+                                    "",
                                     {output_path},
                                     {},
                                     {},
@@ -309,27 +342,29 @@ inline void SetLauncher() {
         action->SetCacheFlag(IExecutionAction::CacheFlag::CacheOutput);
 
         // run execution
-        auto response = action->Execute();
+        auto const response = action->Execute();
         REQUIRE(response);
 
         // verify result
         CHECK(response->ExitCode() == 1);
-        auto artifacts = response->Artifacts();
-        REQUIRE(artifacts.contains(output_path));
-        CHECK(artifacts.at(output_path).digest == test_digest);
+        auto const artifacts = response->Artifacts();
+        REQUIRE(artifacts.has_value());
+        REQUIRE(artifacts.value()->contains(output_path));
+        CHECK(artifacts.value()->at(output_path).digest == test_digest);
 
         CHECK(not response->IsCached());
 
         SECTION("Rerun execution to verify that non-zero actions are rerun") {
             // run execution
-            auto response = action->Execute();
+            auto const response = action->Execute();
             REQUIRE(response);
 
             // verify result
             CHECK(response->ExitCode() == 1);
-            auto artifacts = response->Artifacts();
-            REQUIRE(artifacts.contains(output_path));
-            CHECK(artifacts.at(output_path).digest == test_digest);
+            auto const artifacts = response->Artifacts();
+            REQUIRE(artifacts.has_value());
+            REQUIRE(artifacts.value()->contains(output_path));
+            CHECK(artifacts.value()->at(output_path).digest == test_digest);
             CHECK(not response->IsCached());
         }
     }
@@ -338,14 +373,15 @@ inline void SetLauncher() {
         action->SetCacheFlag(IExecutionAction::CacheFlag::DoNotCacheOutput);
 
         // run execution
-        auto response = action->Execute();
+        auto const response = action->Execute();
         REQUIRE(response);
 
         // verify result
         CHECK(response->ExitCode() == 1);
-        auto artifacts = response->Artifacts();
-        REQUIRE(artifacts.contains(output_path));
-        CHECK(artifacts.at(output_path).digest == test_digest);
+        auto const artifacts = response->Artifacts();
+        REQUIRE(artifacts.has_value());
+        REQUIRE(artifacts.value()->contains(output_path));
+        CHECK(artifacts.value()->at(output_path).digest == test_digest);
         CHECK(not response->IsCached());
 
         SECTION("Rerun execution to verify non-zero actions are not cached") {
@@ -355,9 +391,10 @@ inline void SetLauncher() {
 
             // verify result
             CHECK(response->ExitCode() == 1);
-            auto artifacts = response->Artifacts();
-            REQUIRE(artifacts.contains(output_path));
-            CHECK(artifacts.at(output_path).digest == test_digest);
+            auto const artifacts = response->Artifacts();
+            REQUIRE(artifacts.has_value());
+            REQUIRE(artifacts.value()->contains(output_path));
+            CHECK(artifacts.value()->at(output_path).digest == test_digest);
             CHECK(not response->IsCached());
         }
     }
@@ -392,6 +429,7 @@ inline void SetLauncher() {
 
     auto action = api->CreateAction(*api->UploadTree({}),
                                     {"/bin/sh", "-c", make_cmd("root")},
+                                    "",
                                     {},
                                     {"root"},
                                     env,
@@ -400,7 +438,7 @@ inline void SetLauncher() {
     action->SetCacheFlag(IExecutionAction::CacheFlag::CacheOutput);
 
     // run execution
-    auto response = action->Execute();
+    auto const response = action->Execute();
     REQUIRE(response);
 
     // verify result
@@ -410,10 +448,11 @@ inline void SetLauncher() {
         CHECK_FALSE(response->IsCached());
     }
 
-    auto artifacts = response->Artifacts();
-    REQUIRE_FALSE(artifacts.empty());
+    auto const artifacts = response->Artifacts();
+    REQUIRE(artifacts.has_value());
+    REQUIRE_FALSE(artifacts.value()->empty());
 
-    auto info = artifacts.begin()->second;
+    auto info = artifacts.value()->begin()->second;
 
     SECTION("retrieve via same API object") {
         auto out_path = GetTestDir(test_name) / "out1";
@@ -464,6 +503,7 @@ TestRetrieveFileAndSymlinkWithSameContentToPath(ApiFactory const& api_factory,
 
     auto action = api->CreateAction(*api->UploadTree({}),
                                     {"/bin/sh", "-c", make_cmd("root")},
+                                    "",
                                     {},
                                     {"root"},
                                     env,
@@ -472,7 +512,7 @@ TestRetrieveFileAndSymlinkWithSameContentToPath(ApiFactory const& api_factory,
     action->SetCacheFlag(IExecutionAction::CacheFlag::CacheOutput);
 
     // run execution
-    auto response = action->Execute();
+    auto const response = action->Execute();
     REQUIRE(response);
 
     // verify result
@@ -482,10 +522,11 @@ TestRetrieveFileAndSymlinkWithSameContentToPath(ApiFactory const& api_factory,
         CHECK_FALSE(response->IsCached());
     }
 
-    auto artifacts = response->Artifacts();
-    REQUIRE_FALSE(artifacts.empty());
+    auto const artifacts = response->Artifacts();
+    REQUIRE(artifacts.has_value());
+    REQUIRE_FALSE(artifacts.value()->empty());
 
-    auto info = artifacts.begin()->second;
+    auto info = artifacts.value()->begin()->second;
 
     SECTION("retrieve via same API object") {
         auto out_path = GetTestDir(test_name) / "out1";
@@ -532,6 +573,7 @@ TestRetrieveFileAndSymlinkWithSameContentToPath(ApiFactory const& api_factory,
 
     auto action = api->CreateAction(*api->UploadTree({}),
                                     {"/bin/sh", "-c", cmd},
+                                    "",
                                     {foo_path.string(), link_path.string()},
                                     {bar_path.parent_path().string()},
                                     env,
@@ -540,7 +582,7 @@ TestRetrieveFileAndSymlinkWithSameContentToPath(ApiFactory const& api_factory,
     action->SetCacheFlag(IExecutionAction::CacheFlag::CacheOutput);
 
     // run execution
-    auto response = action->Execute();
+    auto const response = action->Execute();
     REQUIRE(response);
 
     // verify result
@@ -550,16 +592,17 @@ TestRetrieveFileAndSymlinkWithSameContentToPath(ApiFactory const& api_factory,
         CHECK_FALSE(response->IsCached());
     }
 
-    auto artifacts = response->Artifacts();
-    REQUIRE_FALSE(artifacts.empty());
+    auto const artifacts = response->Artifacts();
+    REQUIRE(artifacts.has_value());
+    REQUIRE_FALSE(artifacts.value()->empty());
 
     std::vector<std::filesystem::path> paths{};
     std::vector<Artifact::ObjectInfo> infos{};
 
     SECTION("retrieve via same API object") {
         auto out_path = GetTestDir(test_name) / "out1";
-        std::for_each(artifacts.begin(),
-                      artifacts.end(),
+        std::for_each(artifacts.value()->begin(),
+                      artifacts.value()->end(),
                       [&out_path, &paths, &infos](auto const& entry) {
                           paths.emplace_back(out_path / entry.first);
                           infos.emplace_back(entry.second);
@@ -573,8 +616,8 @@ TestRetrieveFileAndSymlinkWithSameContentToPath(ApiFactory const& api_factory,
     SECTION("retrieve from new API object but same endpoint") {
         auto second_api = api_factory();
         auto out_path = GetTestDir(test_name) / "out2";
-        std::for_each(artifacts.begin(),
-                      artifacts.end(),
+        std::for_each(artifacts.value()->begin(),
+                      artifacts.value()->end(),
                       [&out_path, &paths, &infos](auto const& entry) {
                           paths.emplace_back(out_path / entry.first);
                           infos.emplace_back(entry.second);
@@ -599,6 +642,7 @@ TestRetrieveFileAndSymlinkWithSameContentToPath(ApiFactory const& api_factory,
         {"/bin/sh",
          "-c",
          fmt::format("set -e\n [ -d {} ]", output_path.string())},
+        "",
         {},
         {output_path},
         {},
@@ -608,26 +652,28 @@ TestRetrieveFileAndSymlinkWithSameContentToPath(ApiFactory const& api_factory,
         action->SetCacheFlag(IExecutionAction::CacheFlag::CacheOutput);
 
         // run execution
-        auto response = action->Execute();
+        auto const response = action->Execute();
         REQUIRE(response);
 
         // verify result
-        auto artifacts = response->Artifacts();
-        REQUIRE(artifacts.contains(output_path));
-        CHECK(IsTreeObject(artifacts.at(output_path).type));
+        auto const artifacts = response->Artifacts();
+        REQUIRE(artifacts.has_value());
+        REQUIRE(artifacts.value()->contains(output_path));
+        CHECK(IsTreeObject(artifacts.value()->at(output_path).type));
 
         if (is_hermetic) {
             CHECK(not response->IsCached());
 
             SECTION("Rerun execution to verify caching") {
                 // run execution
-                auto response = action->Execute();
+                auto const response = action->Execute();
                 REQUIRE(response);
 
                 // verify result
-                auto artifacts = response->Artifacts();
-                REQUIRE(artifacts.contains(output_path));
-                CHECK(IsTreeObject(artifacts.at(output_path).type));
+                auto const artifacts = response->Artifacts();
+                REQUIRE(artifacts.has_value());
+                REQUIRE(artifacts.value()->contains(output_path));
+                CHECK(IsTreeObject(artifacts.value()->at(output_path).type));
                 CHECK(response->IsCached());
             }
         }
@@ -637,25 +683,29 @@ TestRetrieveFileAndSymlinkWithSameContentToPath(ApiFactory const& api_factory,
         action->SetCacheFlag(IExecutionAction::CacheFlag::DoNotCacheOutput);
 
         // run execution
-        auto response = action->Execute();
+        auto const response = action->Execute();
         REQUIRE(response);
 
         // verify result
-        auto artifacts = response->Artifacts();
-        REQUIRE(artifacts.contains(output_path));
-        CHECK(IsTreeObject(artifacts.at(output_path).type));
+        auto const artifacts = response->Artifacts();
+        REQUIRE(artifacts.has_value());
+        REQUIRE(artifacts.value()->contains(output_path));
+        CHECK(IsTreeObject(artifacts.value()->at(output_path).type));
         CHECK(not response->IsCached());
 
         SECTION("Rerun execution to verify caching") {
             // run execution
-            auto response = action->Execute();
+            auto const response = action->Execute();
             REQUIRE(response);
 
             // verify result
-            auto artifacts = response->Artifacts();
-            REQUIRE(artifacts.contains(output_path));
-            CHECK(IsTreeObject(artifacts.at(output_path).type));
+            auto const artifacts = response->Artifacts();
+            REQUIRE(artifacts.has_value());
+            REQUIRE(artifacts.value()->contains(output_path));
+            CHECK(IsTreeObject(artifacts.value()->at(output_path).type));
             CHECK(not response->IsCached());
         }
     }
 }
+
+#endif  // INCLUDED_SRC_TEST_BUILDTOOL_EXECUTION_API_COMMON_API_TEST_HPP
