@@ -12,9 +12,12 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+#include "src/buildtool/execution_engine/executor/executor.hpp"
+
 #include <algorithm>
 #include <filesystem>
 #include <map>
+#include <memory>
 #include <optional>
 #include <string>
 #include <unordered_map>
@@ -23,13 +26,22 @@
 
 #include "catch2/catch_test_macros.hpp"
 #include "gsl/gsl"
-#include "src/buildtool/common/artifact_factory.hpp"
+#include "src/buildtool/auth/authentication.hpp"
+#include "src/buildtool/common/artifact_description.hpp"
+#include "src/buildtool/common/artifact_digest.hpp"
+#include "src/buildtool/common/artifact_digest_factory.hpp"
 #include "src/buildtool/common/repository_config.hpp"
 #include "src/buildtool/common/statistics.hpp"
+#include "src/buildtool/crypto/hash_function.hpp"
 #include "src/buildtool/execution_api/common/execution_api.hpp"
-#include "src/buildtool/execution_engine/executor/executor.hpp"
+#include "src/buildtool/execution_api/remote/config.hpp"
+#include "src/buildtool/execution_api/remote/context.hpp"
+#include "src/buildtool/execution_engine/executor/context.hpp"
 #include "src/buildtool/file_system/file_system_manager.hpp"
 #include "src/buildtool/progress_reporting/progress.hpp"
+#include "src/utils/cpp/expected.hpp"
+#include "test/utils/executor/test_api_bundle.hpp"
+#include "test/utils/hermeticity/test_hash_function_type.hpp"
 
 /// \brief Mockup API test config.
 struct TestApiConfig {
@@ -40,7 +52,7 @@ struct TestApiConfig {
 
     struct TestExecutionConfig {
         bool failed{};
-        std::vector<std::string> outputs{};
+        std::vector<std::string> outputs;
     };
 
     struct TestResponseConfig {
@@ -48,11 +60,17 @@ struct TestApiConfig {
         int exit_code{};
     };
 
-    std::unordered_map<std::string, TestArtifactConfig> artifacts{};
+    std::unordered_map<std::string, TestArtifactConfig> artifacts;
     TestExecutionConfig execution;
     TestResponseConfig response;
 };
 
+static auto NamedDigest(std::string const& str) -> ArtifactDigest {
+    HashFunction const hash_function{TestHashType::ReadFromEnvironment()};
+    return ArtifactDigestFactory::HashDataAs<ObjectType::File>(hash_function,
+                                                               str);
+}
+
 // forward declarations
 class TestApi;
 class TestAction;
@@ -60,9 +78,10 @@ class TestResponse;
 
 /// \brief Mockup Response, stores only config and action result
 class TestResponse : public IExecutionResponse {
-    friend class TestAction;
-
   public:
+    explicit TestResponse(TestApiConfig config) noexcept
+        : config_{std::move(config)} {}
+
     [[nodiscard]] auto Status() const noexcept -> StatusCode final {
         return StatusCode::Success;
     }
@@ -76,10 +95,35 @@ class TestResponse : public IExecutionResponse {
     [[nodiscard]] auto HasStdOut() const noexcept -> bool final { return true; }
     [[nodiscard]] auto StdErr() noexcept -> std::string final { return {}; }
     [[nodiscard]] auto StdOut() noexcept -> std::string final { return {}; }
-    [[nodiscard]] auto ActionDigest() const noexcept -> std::string final {
-        return {};
+    [[nodiscard]] auto ActionDigest() const noexcept
+        -> std::string const& final {
+        static const std::string kEmptyHash;
+        return kEmptyHash;
     }
-    [[nodiscard]] auto Artifacts() noexcept -> ArtifactInfos final {
+    [[nodiscard]] auto Artifacts() noexcept
+        -> expected<gsl::not_null<ArtifactInfos const*>, std::string> final {
+        if (not populated_) {
+            Populate();
+        }
+        return gsl::not_null<ArtifactInfos const*>(&artifacts_);
+    }
+    [[nodiscard]] auto DirectorySymlinks() noexcept
+        -> expected<gsl::not_null<DirSymlinks const*>, std::string> final {
+        static const DirSymlinks kEmptySymlinks{};
+        return gsl::not_null<DirSymlinks const*>(&kEmptySymlinks);
+    }
+
+  private:
+    TestApiConfig config_{};
+    ArtifactInfos artifacts_;
+    bool populated_ = false;
+
+    void Populate() noexcept {
+        if (populated_) {
+            return;
+        }
+        populated_ = true;
+
         ArtifactInfos artifacts{};
         artifacts.reserve(config_.execution.outputs.size());
 
@@ -88,46 +132,34 @@ class TestResponse : public IExecutionResponse {
             try {
                 artifacts.emplace(
                     path,
-                    Artifact::ObjectInfo{
-                        .digest = ArtifactDigest{path, 0, /*is_tree=*/false},
-                        .type = ObjectType::File});
+                    Artifact::ObjectInfo{.digest = NamedDigest(path),
+                                         .type = ObjectType::File});
             } catch (...) {
-                return {};
+                return;
             }
         }
-
-        return artifacts;
+        artifacts_ = std::move(artifacts);
     }
-    [[nodiscard]] auto ArtifactsWithDirSymlinks() noexcept
-        -> std::pair<ArtifactInfos, DirSymlinks> final {
-        return {};
-    }
-
-  private:
-    TestApiConfig config_{};
-    explicit TestResponse(TestApiConfig config) noexcept
-        : config_{std::move(config)} {}
 };
 
 /// \brief Mockup Action, stores only config
 class TestAction : public IExecutionAction {
-    friend class TestApi;
-
   public:
+    explicit TestAction(TestApiConfig config) noexcept
+        : config_{std::move(config)} {}
+
     auto Execute(Logger const* /*unused*/) noexcept
         -> IExecutionResponse::Ptr final {
         if (config_.execution.failed) {
             return nullptr;
         }
-        return IExecutionResponse::Ptr{new TestResponse{config_}};
+        return std::make_unique<TestResponse>(config_);
     }
     void SetCacheFlag(CacheFlag /*unused*/) noexcept final {}
     void SetTimeout(std::chrono::milliseconds /*unused*/) noexcept final {}
 
   private:
     TestApiConfig config_{};
-    explicit TestAction(TestApiConfig config) noexcept
-        : config_{std::move(config)} {}
 };
 
 /// \brief Mockup Api, use config to create action and handle artifact upload
@@ -136,32 +168,32 @@ class TestApi : public IExecutionApi {
     explicit TestApi(TestApiConfig config) noexcept
         : config_{std::move(config)} {}
 
-    auto CreateAction(
+    [[nodiscard]] auto CreateAction(
         ArtifactDigest const& /*unused*/,
         std::vector<std::string> const& /*unused*/,
+        std::string const& /*unused*/,
         std::vector<std::string> const& /*unused*/,
         std::vector<std::string> const& /*unused*/,
         std::map<std::string, std::string> const& /*unused*/,
-        std::map<std::string, std::string> const& /*unused*/) noexcept
+        std::map<std::string, std::string> const& /*unused*/) const noexcept
         -> IExecutionAction::Ptr final {
-        return IExecutionAction::Ptr{new TestAction(config_)};
+        return std::make_unique<TestAction>(config_);
     }
-    auto RetrieveToPaths(
+    [[nodiscard]] auto RetrieveToPaths(
         std::vector<Artifact::ObjectInfo> const& /*unused*/,
         std::vector<std::filesystem::path> const& /*unused*/,
-        std::optional<
-            gsl::not_null<IExecutionApi*>> const& /* unused */) noexcept
-        -> bool final {
+        IExecutionApi const* /* unused */) const noexcept -> bool final {
         return false;  // not needed by Executor
     }
-    auto RetrieveToFds(std::vector<Artifact::ObjectInfo> const& /*unused*/,
-                       std::vector<int> const& /*unused*/,
-                       bool /*unused*/) noexcept -> bool final {
+    [[nodiscard]] auto RetrieveToFds(
+        std::vector<Artifact::ObjectInfo> const& /*unused*/,
+        std::vector<int> const& /*unused*/,
+        bool /*unused*/) const noexcept -> bool final {
         return false;  // not needed by Executor
     }
-    auto RetrieveToCas(std::vector<Artifact::ObjectInfo> const& unused,
-                       gsl::not_null<IExecutionApi*> const& /*unused*/) noexcept
-        -> bool final {
+    [[nodiscard]] auto RetrieveToCas(
+        std::vector<Artifact::ObjectInfo> const& unused,
+        IExecutionApi const& /*unused*/) const noexcept -> bool final {
         // Note that a false-positive "free-nonheap-object" warning is thrown by
         // gcc 12.2 with GNU libstdc++, if the caller passes a temporary vector
         // that is not used by this function. Therefore, we explicitly use this
@@ -169,25 +201,29 @@ class TestApi : public IExecutionApi {
         // irrelevant for testing though.
         return unused.empty();  // not needed by Executor
     }
-    auto RetrieveToMemory(
-        Artifact::ObjectInfo const& /*artifact_info*/) noexcept
+    [[nodiscard]] auto RetrieveToMemory(
+        Artifact::ObjectInfo const& /*artifact_info*/) const noexcept
         -> std::optional<std::string> override {
         return std::nullopt;  // not needed by Executor
     }
-    auto Upload(BlobContainer const& blobs, bool /*unused*/) noexcept
-        -> bool final {
+    [[nodiscard]] auto Upload(ArtifactBlobContainer&& blobs,
+                              bool /*unused*/) const noexcept -> bool final {
+        auto blob_range = blobs.Blobs();
         return std::all_of(
-            blobs.begin(), blobs.end(), [this](auto const& blob) {
-                return config_.artifacts[blob.data]
-                           .uploads  // for local artifacts
-                       or config_.artifacts[blob.digest.hash()]
-                              .uploads;  // for known and action artifacts
+            blob_range.begin(), blob_range.end(), [this](auto const& blob) {
+                // for local artifacts
+                auto it1 = config_.artifacts.find(*blob.data);
+                if (it1 != config_.artifacts.end() and it1->second.uploads) {
+                    return true;
+                }
+                // for known and action artifacts
+                auto it2 = config_.artifacts.find(blob.digest.hash());
+                return it2 != config_.artifacts.end() and it2->second.uploads;
             });
     }
-    auto UploadTree(
-        std::vector<
-            DependencyGraph::NamedArtifactNodePtr> const& /*unused*/) noexcept
-        -> std::optional<ArtifactDigest> final {
+    [[nodiscard]] auto UploadTree(
+        std::vector<DependencyGraph::NamedArtifactNodePtr> const& /*unused*/)
+        const noexcept -> std::optional<ArtifactDigest> final {
         return ArtifactDigest{};  // not needed by Executor
     }
     [[nodiscard]] auto IsAvailable(ArtifactDigest const& digest) const noexcept
@@ -230,25 +266,27 @@ class TestApi : public IExecutionApi {
     -> std::pair<TestApiConfig, RepositoryConfig> {
     using path = std::filesystem::path;
 
-    auto const local_cpp_desc = ArtifactDescription{path{"local.cpp"}, ""};
-    auto const known_cpp_desc = ArtifactDescription{
-        ArtifactDigest{"known.cpp", 0, /*is_tree=*/false}, ObjectType::File};
+    auto const local_cpp_desc =
+        ArtifactDescription::CreateLocal(path{"local.cpp"}, "");
+    auto const known_digest = NamedDigest("known.cpp");
+    auto const known_cpp_desc =
+        ArtifactDescription::CreateKnown(known_digest, ObjectType::File);
 
     auto const test_action_desc = ActionDescription{
         {"output1.exe", "output2.exe"},
         {},
         Action{"test_action", {"cmd", "line"}, {}},
-        {{"local.cpp", local_cpp_desc}, {"known.cpp", known_cpp_desc}}};
+        {{"local.cpp", local_cpp_desc}, {known_digest.hash(), known_cpp_desc}}};
 
     CHECK(g->AddAction(test_action_desc));
     CHECK(FileSystemManager::WriteFile("local.cpp", ws / "local.cpp"));
 
     TestApiConfig config{};
 
-    config.artifacts["local.cpp"].uploads = true;
-    config.artifacts["known.cpp"].available = true;
-    config.artifacts["output1.exe"].available = true;
-    config.artifacts["output2.exe"].available = true;
+    config.artifacts[NamedDigest("local.cpp").hash()].uploads = true;
+    config.artifacts[NamedDigest("known.cpp").hash()].available = true;
+    config.artifacts[NamedDigest("output1.exe").hash()].available = true;
+    config.artifacts[NamedDigest("output2.exe").hash()].available = true;
 
     config.execution.failed = false;
     config.execution.outputs = {"output1.exe", "output2.exe"};
@@ -265,46 +303,69 @@ TEST_CASE("Executor: Process artifact", "[executor]") {
     DependencyGraph g;
     auto [config, repo_config] = CreateTest(&g, workspace_path);
 
-    auto const local_cpp_desc =
-        ArtifactFactory::DescribeLocalArtifact("local.cpp", "");
-    auto const local_cpp_id = ArtifactFactory::Identifier(local_cpp_desc);
+    HashFunction const hash_function{TestHashType::ReadFromEnvironment()};
 
-    auto const known_cpp_desc = ArtifactFactory::DescribeKnownArtifact(
-        "known.cpp", 0, ObjectType::File);
-    auto const known_cpp_id = ArtifactFactory::Identifier(known_cpp_desc);
+    auto const local_cpp_id =
+        ArtifactDescription::CreateLocal("local.cpp", "").Id();
+
+    auto const known_cpp_id = ArtifactDescription::CreateKnown(
+                                  NamedDigest("known.cpp"), ObjectType::File)
+                                  .Id();
+
+    Auth auth{};
+    RetryConfig retry_config{};             // default retry config
+    RemoteExecutionConfig remote_config{};  // default remote config
+    RemoteContext const remote_context{.auth = &auth,
+                                       .retry_config = &retry_config,
+                                       .exec_config = &remote_config};
 
     SECTION("Processing succeeds for valid config") {
-        auto api = TestApi::Ptr{new TestApi{config}};
+        auto api = std::make_shared<TestApi>(config);
         Statistics stats{};
         Progress progress{};
-        Executor runner{
-            &repo_config, api.get(), api.get(), {}, {}, &stats, &progress};
+        auto const apis = CreateTestApiBundle(&hash_function, api);
+        ExecutionContext const exec_context{.repo_config = &repo_config,
+                                            .apis = &apis,
+                                            .remote_context = &remote_context,
+                                            .statistics = &stats,
+                                            .progress = &progress};
+        Executor runner{&exec_context};
 
         CHECK(runner.Process(g.ArtifactNodeWithId(local_cpp_id)));
         CHECK(runner.Process(g.ArtifactNodeWithId(known_cpp_id)));
     }
 
     SECTION("Processing fails if uploading local artifact failed") {
-        config.artifacts["local.cpp"].uploads = false;
+        config.artifacts[NamedDigest("local.cpp").hash()].uploads = false;
 
-        auto api = TestApi::Ptr{new TestApi{config}};
+        auto api = std::make_shared<TestApi>(config);
         Statistics stats{};
         Progress progress{};
-        Executor runner{
-            &repo_config, api.get(), api.get(), {}, {}, &stats, &progress};
+        auto const apis = CreateTestApiBundle(&hash_function, api);
+        ExecutionContext const exec_context{.repo_config = &repo_config,
+                                            .apis = &apis,
+                                            .remote_context = &remote_context,
+                                            .statistics = &stats,
+                                            .progress = &progress};
+        Executor runner{&exec_context};
 
         CHECK(not runner.Process(g.ArtifactNodeWithId(local_cpp_id)));
         CHECK(runner.Process(g.ArtifactNodeWithId(known_cpp_id)));
     }
 
     SECTION("Processing fails if known artifact is not available") {
-        config.artifacts["known.cpp"].available = false;
+        config.artifacts[NamedDigest("known.cpp").hash()].available = false;
 
-        auto api = TestApi::Ptr{new TestApi{config}};
+        auto api = std::make_shared<TestApi>(config);
         Statistics stats{};
         Progress progress{};
-        Executor runner{
-            &repo_config, api.get(), api.get(), {}, {}, &stats, &progress};
+        auto const apis = CreateTestApiBundle(&hash_function, api);
+        ExecutionContext const exec_context{.repo_config = &repo_config,
+                                            .apis = &apis,
+                                            .remote_context = &remote_context,
+                                            .statistics = &stats,
+                                            .progress = &progress};
+        Executor runner{&exec_context};
 
         CHECK(runner.Process(g.ArtifactNodeWithId(local_cpp_id)));
         CHECK(not runner.Process(g.ArtifactNodeWithId(known_cpp_id)));
@@ -318,29 +379,40 @@ TEST_CASE("Executor: Process action", "[executor]") {
     DependencyGraph g;
     auto [config, repo_config] = CreateTest(&g, workspace_path);
 
-    auto const local_cpp_desc =
-        ArtifactFactory::DescribeLocalArtifact("local.cpp", "");
-    auto const local_cpp_id = ArtifactFactory::Identifier(local_cpp_desc);
+    HashFunction const hash_function{TestHashType::ReadFromEnvironment()};
 
-    auto const known_cpp_desc = ArtifactFactory::DescribeKnownArtifact(
-        "known.cpp", 0, ObjectType::File);
-    auto const known_cpp_id = ArtifactFactory::Identifier(known_cpp_desc);
+    auto const local_cpp_id =
+        ArtifactDescription::CreateLocal("local.cpp", "").Id();
 
-    ActionIdentifier action_id{"test_action"};
-    auto const output1_desc =
-        ArtifactFactory::DescribeActionArtifact(action_id, "output1.exe");
-    auto const output1_id = ArtifactFactory::Identifier(output1_desc);
+    auto const known_cpp_id = ArtifactDescription::CreateKnown(
+                                  NamedDigest("known.cpp"), ObjectType::File)
+                                  .Id();
 
-    auto const output2_desc =
-        ArtifactFactory::DescribeActionArtifact(action_id, "output2.exe");
-    auto const output2_id = ArtifactFactory::Identifier(output2_desc);
+    ActionIdentifier const action_id{"test_action"};
+    auto const output1_id =
+        ArtifactDescription::CreateAction(action_id, "output1.exe").Id();
+
+    auto const output2_id =
+        ArtifactDescription::CreateAction(action_id, "output2.exe").Id();
+
+    Auth auth{};
+    RetryConfig retry_config{};             // default retry config
+    RemoteExecutionConfig remote_config{};  // default remote config
+    RemoteContext const remote_context{.auth = &auth,
+                                       .retry_config = &retry_config,
+                                       .exec_config = &remote_config};
 
     SECTION("Processing succeeds for valid config") {
-        auto api = TestApi::Ptr{new TestApi{config}};
+        auto api = std::make_shared<TestApi>(config);
         Statistics stats{};
         Progress progress{};
-        Executor runner{
-            &repo_config, api.get(), api.get(), {}, {}, &stats, &progress};
+        auto const apis = CreateTestApiBundle(&hash_function, api);
+        ExecutionContext const exec_context{.repo_config = &repo_config,
+                                            .apis = &apis,
+                                            .remote_context = &remote_context,
+                                            .statistics = &stats,
+                                            .progress = &progress};
+        Executor runner{&exec_context};
 
         CHECK(runner.Process(g.ArtifactNodeWithId(local_cpp_id)));
         CHECK(runner.Process(g.ArtifactNodeWithId(known_cpp_id)));
@@ -352,11 +424,16 @@ TEST_CASE("Executor: Process action", "[executor]") {
     SECTION("Processing succeeds even if result was is not cached") {
         config.response.cached = false;
 
-        auto api = TestApi::Ptr{new TestApi{config}};
+        auto api = std::make_shared<TestApi>(config);
         Statistics stats{};
         Progress progress{};
-        Executor runner{
-            &repo_config, api.get(), api.get(), {}, {}, &stats, &progress};
+        auto const apis = CreateTestApiBundle(&hash_function, api);
+        ExecutionContext const exec_context{.repo_config = &repo_config,
+                                            .apis = &apis,
+                                            .remote_context = &remote_context,
+                                            .statistics = &stats,
+                                            .progress = &progress};
+        Executor runner{&exec_context};
 
         CHECK(runner.Process(g.ArtifactNodeWithId(local_cpp_id)));
         CHECK(runner.Process(g.ArtifactNodeWithId(known_cpp_id)));
@@ -366,13 +443,18 @@ TEST_CASE("Executor: Process action", "[executor]") {
     }
 
     SECTION("Processing succeeds even if output is not available in CAS") {
-        config.artifacts["output2.exe"].available = false;
+        config.artifacts[NamedDigest("output2.exe").hash()].available = false;
 
-        auto api = TestApi::Ptr{new TestApi{config}};
+        auto api = std::make_shared<TestApi>(config);
         Statistics stats{};
         Progress progress{};
-        Executor runner{
-            &repo_config, api.get(), api.get(), {}, {}, &stats, &progress};
+        auto const apis = CreateTestApiBundle(&hash_function, api);
+        ExecutionContext const exec_context{.repo_config = &repo_config,
+                                            .apis = &apis,
+                                            .remote_context = &remote_context,
+                                            .statistics = &stats,
+                                            .progress = &progress};
+        Executor runner{&exec_context};
 
         CHECK(runner.Process(g.ArtifactNodeWithId(local_cpp_id)));
         CHECK(runner.Process(g.ArtifactNodeWithId(known_cpp_id)));
@@ -387,11 +469,16 @@ TEST_CASE("Executor: Process action", "[executor]") {
     SECTION("Processing fails if execution failed") {
         config.execution.failed = true;
 
-        auto api = TestApi::Ptr{new TestApi{config}};
+        auto api = std::make_shared<TestApi>(config);
         Statistics stats{};
         Progress progress{};
-        Executor runner{
-            &repo_config, api.get(), api.get(), {}, {}, &stats, &progress};
+        auto const apis = CreateTestApiBundle(&hash_function, api);
+        ExecutionContext const exec_context{.repo_config = &repo_config,
+                                            .apis = &apis,
+                                            .remote_context = &remote_context,
+                                            .statistics = &stats,
+                                            .progress = &progress};
+        Executor runner{&exec_context};
 
         CHECK(runner.Process(g.ArtifactNodeWithId(local_cpp_id)));
         CHECK(runner.Process(g.ArtifactNodeWithId(known_cpp_id)));
@@ -403,11 +490,16 @@ TEST_CASE("Executor: Process action", "[executor]") {
     SECTION("Processing fails if exit code is non-zero") {
         config.response.exit_code = 1;
 
-        auto api = TestApi::Ptr{new TestApi{config}};
+        auto api = std::make_shared<TestApi>(config);
         Statistics stats{};
         Progress progress{};
-        Executor runner{
-            &repo_config, api.get(), api.get(), {}, {}, &stats, &progress};
+        auto const apis = CreateTestApiBundle(&hash_function, api);
+        ExecutionContext const exec_context{.repo_config = &repo_config,
+                                            .apis = &apis,
+                                            .remote_context = &remote_context,
+                                            .statistics = &stats,
+                                            .progress = &progress};
+        Executor runner{&exec_context};
 
         CHECK(runner.Process(g.ArtifactNodeWithId(local_cpp_id)));
         CHECK(runner.Process(g.ArtifactNodeWithId(known_cpp_id)));
@@ -422,11 +514,16 @@ TEST_CASE("Executor: Process action", "[executor]") {
     SECTION("Processing fails if any output is missing") {
         config.execution.outputs = {"output1.exe" /*, "output2.exe"*/};
 
-        auto api = TestApi::Ptr{new TestApi{config}};
+        auto api = std::make_shared<TestApi>(config);
         Statistics stats{};
         Progress progress{};
-        Executor runner{
-            &repo_config, api.get(), api.get(), {}, {}, &stats, &progress};
+        auto const apis = CreateTestApiBundle(&hash_function, api);
+        ExecutionContext const exec_context{.repo_config = &repo_config,
+                                            .apis = &apis,
+                                            .remote_context = &remote_context,
+                                            .statistics = &stats,
+                                            .progress = &progress};
+        Executor runner{&exec_context};
 
         CHECK(runner.Process(g.ArtifactNodeWithId(local_cpp_id)));
         CHECK(runner.Process(g.ArtifactNodeWithId(known_cpp_id)));
