@@ -12,6 +12,8 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+#include "src/buildtool/file_system/git_repo.hpp"
+
 #include <atomic>
 #include <cstdlib>
 #include <filesystem>
@@ -19,7 +21,6 @@
 #include <optional>
 #include <string>
 #include <thread>
-#include <variant>
 #include <vector>
 
 #include "catch2/catch_test_macros.hpp"
@@ -27,11 +28,12 @@
 #include "nlohmann/json.hpp"
 #include "src/buildtool/file_system/file_system_manager.hpp"
 #include "src/buildtool/file_system/git_cas.hpp"
-#include "src/buildtool/file_system/git_repo.hpp"
 #include "src/buildtool/logging/log_level.hpp"
 #include "src/buildtool/logging/logger.hpp"
+#include "src/buildtool/storage/config.hpp"
 #include "src/utils/cpp/atomic.hpp"
 #include "src/utils/cpp/hex_string.hpp"
+#include "test/utils/hermeticity/test_storage_config.hpp"
 #include "test/utils/shell_quoting.hpp"
 
 namespace {
@@ -76,7 +78,7 @@ class TestUtils {
                                               : (*repo_path / ".git").string()),
                         QuoteForShell(repo_path->string()));
         if (std::system(cmd.c_str()) == 0) {
-            return *repo_path;
+            return repo_path;
         }
         return std::nullopt;
     }
@@ -181,7 +183,7 @@ TEST_CASE("Single-threaded real repository local operations", "[git_repo]") {
                         std::string(msg));
         });
 
-    SECTION("Stage and commit all") {
+    SECTION("Commit directory") {
         // make blank repo
         auto repo_commit_path = TestUtils::GetRepoPath();
         auto repo_commit =
@@ -195,9 +197,9 @@ TEST_CASE("Single-threaded real repository local operations", "[git_repo]") {
         REQUIRE(FileSystemManager::WriteFile(
             "test no 2", repo_commit_path / "test2.txt", true));
 
-        // stage and commit all
-        auto commit =
-            repo_commit->StageAndCommitAllAnonymous("test commit", logger);
+        // commit subdir
+        auto commit = repo_commit->CommitDirectory(
+            repo_commit_path, "test commit", logger);
         CHECK(commit);
     }
 
@@ -212,12 +214,12 @@ TEST_CASE("Single-threaded real repository local operations", "[git_repo]") {
     }
 
     SECTION("Get head commit") {
-        auto repo_wHead_path = TestUtils::CreateTestRepoWithCheckout();
-        REQUIRE(repo_wHead_path);
-        auto repo_wHead = GitRepo::Open(*repo_wHead_path);
-        REQUIRE(repo_wHead);
+        auto repo_head_path = TestUtils::CreateTestRepoWithCheckout();
+        REQUIRE(repo_head_path);
+        auto repo_head = GitRepo::Open(*repo_head_path);
+        REQUIRE(repo_head);
 
-        auto head_commit = repo_wHead->GetHeadCommit(logger);
+        auto head_commit = repo_head->GetHeadCommit(logger);
         REQUIRE(head_commit);
         CHECK(*head_commit == kRootCommit);
     }
@@ -261,9 +263,9 @@ TEST_CASE("Single-threaded real repository local operations", "[git_repo]") {
         // tag uncommitted tree
         auto foo_bar = GitRepo::tree_entries_t{
             {FromHexString(kFooId).value_or<std::string>({}),
-             {GitRepo::tree_entry_t{"foo", ObjectType::File}}},
+             {GitRepo::TreeEntry{"foo", ObjectType::File}}},
             {FromHexString(kBarId).value_or<std::string>({}),
-             {GitRepo::tree_entry_t{"bar", ObjectType::Executable}}}};
+             {GitRepo::TreeEntry{"bar", ObjectType::Executable}}}};
         auto foo_bar_id = repo_tag->CreateTree(foo_bar);
         REQUIRE(foo_bar_id);
         auto tree_id = ToHexString(*foo_bar_id);
@@ -277,6 +279,8 @@ TEST_CASE("Single-threaded real repository local operations", "[git_repo]") {
 // warning
 TEST_CASE("Single-threaded fake repository operations -- batch 1",
           "[git_repo]") {
+    auto const storage_config = TestStorageConfig::Create();
+
     auto repo_path = TestUtils::CreateTestRepoWithCheckout();
     REQUIRE(repo_path);
     auto cas = GitCAS::Open(*repo_path);
@@ -419,7 +423,7 @@ TEST_CASE("Single-threaded fake repository operations -- batch 1",
 
             // fetch all with base refspecs
             REQUIRE(repo_fetch_all->LocalFetchViaTmpRepo(
-                *repo_path, std::nullopt, logger));
+                storage_config.Get(), *repo_path, std::nullopt, logger));
 
             // check commit is there after fetch
             CHECK(*repo_fetch_all->CheckCommitExists(kRootCommit, logger));
@@ -438,7 +442,7 @@ TEST_CASE("Single-threaded fake repository operations -- batch 1",
 
             // fetch branch
             REQUIRE(repo_fetch_branch->LocalFetchViaTmpRepo(
-                *repo_path, "master", logger));
+                storage_config.Get(), *repo_path, "master", logger));
 
             // check commit is there after fetch
             CHECK(*repo_fetch_branch->CheckCommitExists(kRootCommit, logger));
@@ -468,15 +472,15 @@ TEST_CASE("Single-threaded fake repository operations -- batch 2",
         SECTION("Get base tree id") {
             auto entry_root_c =
                 repo->GetSubtreeFromCommit(kRootCommit, ".", logger);
-            REQUIRE(std::holds_alternative<std::string>(entry_root_c));
-            CHECK(std::get<std::string>(entry_root_c) == kRootId);
+            REQUIRE(entry_root_c);
+            CHECK(*entry_root_c == kRootId);
         }
 
         SECTION("Get inner tree id") {
             auto entry_baz_c =
                 repo->GetSubtreeFromCommit(kRootCommit, "baz", logger);
-            REQUIRE(std::holds_alternative<std::string>(entry_baz_c));
-            CHECK(std::get<std::string>(entry_baz_c) == kBazId);
+            REQUIRE(entry_baz_c);
+            CHECK(*entry_baz_c == kBazId);
         }
     }
 
@@ -580,6 +584,8 @@ TEST_CASE("Single-threaded fake repository operations -- batch 2",
 }
 
 TEST_CASE("Multi-threaded fake repository operations", "[git_repo]") {
+    auto const storage_config = TestStorageConfig::Create();
+
     /*
     Test all fake repository operations while being done in parallel.
     They are supposed to be thread-safe, so no conflicts should exist.
@@ -605,14 +611,17 @@ TEST_CASE("Multi-threaded fake repository operations", "[git_repo]") {
     threads.reserve(kNumThreads);
 
     SECTION("Lookups in the same ODB") {
-        constexpr int NUM_CASES = 10;
+        constexpr int kNumCases = 10;
         for (int id{}; id < kNumThreads; ++id) {
             threads.emplace_back(
-                [&remote_cas, &remote_repo_path, &logger, &starting_signal](
-                    int tid) {
+                [&storage_config,
+                 &remote_cas,
+                 &remote_repo_path,
+                 &logger,
+                 &starting_signal](int tid) {
                     starting_signal.wait(false);
                     // cases based on thread number
-                    switch (tid % NUM_CASES) {
+                    switch (tid % kNumCases) {
                         case 0: {
                             auto remote_repo = GitRepo::Open(remote_cas);
                             REQUIRE(remote_repo);
@@ -621,9 +630,8 @@ TEST_CASE("Multi-threaded fake repository operations", "[git_repo]") {
                             auto entry_baz_c =
                                 remote_repo->GetSubtreeFromCommit(
                                     kRootCommit, "baz", logger);
-                            REQUIRE(std::holds_alternative<std::string>(
-                                entry_baz_c));
-                            CHECK(std::get<std::string>(entry_baz_c) == kBazId);
+                            REQUIRE(entry_baz_c);
+                            CHECK(*entry_baz_c == kBazId);
                         } break;
                         case 1: {
                             auto remote_repo = GitRepo::Open(remote_cas);
@@ -730,8 +738,13 @@ TEST_CASE("Multi-threaded fake repository operations", "[git_repo]") {
                             REQUIRE(remote_repo->IsRepoFake());
                             // fetch all
                             REQUIRE(remote_repo->LocalFetchViaTmpRepo(
-                                *remote_repo_path, std::nullopt, logger));
+                                storage_config.Get(),
+                                *remote_repo_path,
+                                std::nullopt,
+                                logger));
                         } break;
+                        default:
+                            REQUIRE(false);
                     }
                 },
                 id);
