@@ -18,13 +18,11 @@
 
 #include "catch2/catch_session.hpp"
 #include "src/buildtool/file_system/git_context.hpp"
-#include "src/buildtool/storage/config.hpp"
 #include "src/buildtool/storage/file_chunker.hpp"
 #include "test/utils/logging/log_config.hpp"
 
 auto main(int argc, char* argv[]) -> int {
     ConfigureLogging();
-
     /**
      * The current implementation of libgit2 uses pthread_key_t incorrectly
      * on POSIX systems to handle thread-specific data, which requires us to
@@ -33,18 +31,6 @@ auto main(int argc, char* argv[]) -> int {
      */
     GitContext::Create();
 
-    /**
-     * Test must not assume the existence of a home directory, nor write there.
-     * Hence we set the storage root to a fixed location under TEST_TMPDIR which
-     * is set by the test launcher.
-     */
-    auto setup_ok = StorageConfig::SetBuildRoot(
-        std::filesystem::path{std::string{std::getenv("TEST_TMPDIR")}} /
-        ".test_build_root");
-    if (not setup_ok) {
-        return 1;
-    }
-
     // Initialize random content of the file chunker's map.
     FileChunker::Initialize();
 
